/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.context;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A PersistenceContext represents the state of persistent entities
 * @author Lin Liang
 */

public class PersistenceContext {

    /**
     * Map<the id for the entity, the entity>
     */
    private Map<Identity, Object> entities = new HashMap<Identity, Object>();
    
    private Map<Identity, Object> oentities = new HashMap<Identity,Object>();

    private Map<Identity, Integer> hashcodes = new HashMap<Identity, Integer>();
    
    private List orderList = new ArrayList();
   
    /**
     * Map<the id for the entity, the state of the entity>
     */
    private Map<Identity, EntityState> entityStates = new HashMap<Identity, EntityState>();

    private StateManager stateManager = new StateManager();

    public PersistenceContext() {

    }

    /**
     * Add an entity to the context by its identity
     * @param id the identity
     * @param the entity instance
     */
    public void addEntity(Identity id, Object entity, Integer hashcode) {
        entities.put(id, entity);
        hashcodes.put(id, hashcode);
    }
    
    public void addToOrderList(Object entity){
    	orderList.add(entity);
    }
    
    public void clearOrderList(){
    	orderList.clear();
    }
    
    public List getOrderList(){
    	return orderList;
    }
    
    /**
     * add an database from entity to context
     * @param id
     * @param entity
     * @param hashcode
     */
    public void addOEntity(Identity id, Object entity, Integer hashcode) {
        oentities.put(id, entity);
    }
    
    /**
     * remove an database from entity
     * @param id
     */
    public void removeOEntity(Identity id){
    	oentities.remove(id);
    }
    
    /**
     * 
     * @param entity
     */
    public boolean containsO(Object entity){
    	return oentities.containsValue(entity);
    }
    
    public boolean containsByKeyO(Identity id){
    	return oentities.containsKey(id);
    }
    
    /**
     * Remove an entity to the context by its identity
     * @param id the identity
     */
    public void removeEntity(Identity id) {
        entities.remove(id);
        hashcodes.remove(id);
        entityStates.remove(id);
    }

    /**
     * Get the entity instance associated with the given identity
     * @param id the identityof entity
     * @return the entity instance
     */
    public Object getEntity(Identity id) {
        return entities.get(id);
    }

    /**
     * Get the entity instance associated with the given identity
     * @param id the identityof entity
     * @return the entity instance
     */
    public Integer getHashCode(Identity id) {
        return hashcodes.get(id);
    }

    /**
     * Get all the entity instances
     * @return all the entity instances
     */
    @SuppressWarnings("unchecked")
    public List getAllEntities() {
        List allEntities = new ArrayList();
        allEntities.addAll(entities.values());
        return allEntities;
    }
    
    
    /**
     * Get all the identities
     * 
     */
    public List getAllIdentities(){
    	List allIdentities = new ArrayList();
    	allIdentities.addAll(entities.keySet());
    	return allIdentities;
    }
    /**
     * set the state of the entity to context
     * @param entity the entity instance
     * @param state the state of the given entity
     */
    public void setEntityState(Identity id, EntityState state) {
        entityStates.put(id, state);
    }

    /**
     * Retreive the EntityState of the given entity.
     * @param entity The entity
     * @return The EntityState of the given entity.
     */
    public EntityState getEntityState(Identity id) {
        EntityState state = entityStates.get(id);
        if (state == null)
        {
            state = stateManager.getTransientState();
            entityStates.put(id, state);
        }
        return state;
    }

    public StateManager getStateManager() {
        return stateManager;
    }

    /**
     * clear the context
     */
    public void clear() {
        entities.clear();
        entityStates.clear();
    }

    /**
     * check whether the context cache contains the entity
     * @param entity the entity
     * @return true if the context cache contains the entity
     */
    public boolean contains(Object entity) {
        return entities.containsValue(entity);
    }
    public boolean containsByKey(Identity id){
    	return entities.containsKey(id);
    }
    public boolean isTransientState(Identity id){
    	return this.getEntityState(id).equals(stateManager.getTransientState());
    }
    public boolean isCleanState(Identity id){
    	return this.getEntityState(id).equals(stateManager.getCleanState());
    }
    public boolean isDeletedState(Identity id){
    	return this.getEntityState(id).equals(stateManager.getDeletedState());
    }
    public boolean isDirtyState(Identity id){
    	return this.getEntityState(id).equals(stateManager.getDirtyState());
    }
    public boolean isFetchState(Identity id){
    	return this.getEntityState(id).equals(stateManager.getFetchState());
    }
    public boolean isGoneState(Identity id){
    	return this.getEntityState(id).equals(stateManager.getGoneState());
    }
    public boolean isCascadeFindState(Identity id){
    	return this.getEntityState(id).equals(stateManager.getCascadeFindState());
    }
}
