/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.deployer;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.util.List;

import org.ow2.xplus.exception.UnitDeployingException;
import org.ow2.xplus.metadata.PersistenceInfo;

/**
 * Application Manager in persistemce layer. It takes charge of application life
 * cycle management(install, deploy, start, stop and undeploy).
 * @author Lin Liang
 */

public class ApplicationManager {
    /**
     * full path of this unpacked application
     */
    private String appPath = null;

    private PersistenceXMLParser xmlParser = new PersistenceXMLParser();

    /**
     * persistence units infomation
     */
    private List<PersistenceInfo> persistenceInfos = null;

    /**
     * the classes path relative to the persistence.xml path
     */
    private String unitPath;

    public ApplicationManager(String appPath) {
        this.appPath = appPath;
    }

    /**
     * Search a directory for persistence.xml. it should be found in 1) an
     * EJB-JAR file 2) the WEB-INF/classes directory of a WAR file[38] 3) a jar
     * file in the WEB-INF/lib directory of a WAR file 4) a jar file in the root
     * of the EAR 5) a jar file in the EAR library directory 6) an application
     * client jar file
     * @param filePath a String representing the directory to search
     * @return the persistence unit name
     */
    public String parsePersistenceUnit() {

        FileInputStream inputStream = null;

        try {
            if (appPath.endsWith(".jar"))
                handleJar();
            if (appPath.endsWith(".war"))
                handleWar();
            if (appPath.endsWith(".ear"))
                handleEar();
        } catch (IOException ioe) {
            UnitDeployingException ex = new UnitDeployingException(ioe, UnitDeployingException.PROCESSING_UNIT_ERROR);
            ex.showMessage();
        } catch (URISyntaxException ue) {
            UnitDeployingException ex = new UnitDeployingException(ue, UnitDeployingException.PROCESSING_UNIT_ERROR);
            ex.showMessage();
        } finally {
            try {
                if (inputStream != null) {
                    inputStream.close();
                }
            } catch (IOException exc) {
            }
        }
        return null;
    }

    /**
     * Parse a Jar File. persistence.xml should be found in an EJB-JAR file
     * @throws MalformedURLException
     * @throws URISyntaxException
     */
    private void handleJar() throws MalformedURLException, URISyntaxException {
        unitPath = appPath;
        File persistenceXMLFile = new File(unitPath + File.separator + "META-INF" + File.separator + "persistence.xml");
        if (persistenceXMLFile.exists()) {
            parsepersistenceXML(persistenceXMLFile);
        }
    }

    /**
     * Parse a War file. persistence.xml should be found in 1) the
     * WEB-INF/classes directory of a WAR file[38] 2) a jar file in the
     * WEB-INF/lib directory of a WAR file
     * @throws MalformedURLException
     * @throws URISyntaxException
     */
    private void handleWar() throws MalformedURLException, URISyntaxException {
        unitPath = appPath + File.separator + "WEB-INF" + File.separator + "classes";
        File persistenceXMLFile = new File(unitPath + File.separator + "persistence.xml");
        if (!persistenceXMLFile.exists()) {
            unitPath = appPath + File.separator + "WEB-INF" + File.separator + "lib";
            persistenceXMLFile = new File(unitPath + File.separator + "persistence.xml");
        }

        if (persistenceXMLFile.exists()) {
            parsepersistenceXML(persistenceXMLFile);
        }
    }

    /**
     * Parse a Ear file.
     * @throws MalformedURLException
     * @throws URISyntaxException
     */
    private void handleEar() throws MalformedURLException, URISyntaxException {
        unitPath = appPath;
        File persistenceXMLFile = new File(unitPath + File.separator + "META-INF" + File.separator + "persistence.xml");
        if (persistenceXMLFile.exists()) {
            parsepersistenceXML(persistenceXMLFile);
        }
    }

    /**
     * parse persistence.xml to get persistence information
     * @param persistenceXMLFile the persistence.xml file
     */
    private void parsepersistenceXML(File persistenceXMLFile) {
        xmlParser.setUnitPath(unitPath);
        xmlParser.setXmlFile(persistenceXMLFile);
        xmlParser.parse();
        persistenceInfos = xmlParser.getPersistenceInfos();
    }

    /**
     * get all the persistence info of all the units
     * 
     * @return all of the persistence info
     */
    public List<PersistenceInfo> getPersistenceInfos() {
        return persistenceInfos;
    }

}
