/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.deployer;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import org.ow2.xplus.datasvc.JNDIDataSource;
import org.ow2.xplus.datasvc.StandaloneDataSource;
import org.ow2.xplus.exception.UnitDeployingException;
import org.ow2.xplus.logger.LogFactory;
import org.ow2.xplus.logger.Logger;
import org.ow2.xplus.metadata.PersistenceInfo;

/**
 * parse persistence.xml , generate informations
 * @author Lin Liang
 */

public class PersistenceXMLParser {

    public static String PERSISTENCE_NAME = "name";

    public static String PERSISTENCE_TX_TYPE = "transaction-type";

    public static String PERSISTENCE_NON_JTA_DS = "non-jta-data-source";

    public static String PERSISTENCE_JTA_DS = "jta-data-source";

    public static String PERSISTENCE_DB_NAME = "database-name";

    public static String PERSISTENCE_DS_NAME = "data-source-name";

    public static String PERSISTENCE_PROVIDER = "provider";

    public static String PERSISTENCE_CLASS = "class";

    public static String PERSISTENCE_MAPPING_FILE = "mapping-file";

    public static String PERSISTENCE_JAR_FILE = "jar-file";

    public static String PERSISTENCE_EXCLUDE_UNLISTED = "exclude-unlisted-classes";

    public static String PERSISTENCE_PROPERTIES = "properties";

    public static String PERSISTENCE_PROPERTY = "property";

    public static String PERSISTENCE_PROPERTY_NAME = "name";

    public static String PERSISTENCE_PROPERTY_VALUE = "value";

    public static String PERSISTENCE_CONNECTION_PROVIDER = "xplus.connection.provider";

    public static String PERSISTENCE_CONNECTION_DIALECT = "xplus.connection.dialect";

    public static String PERSISTENCE_CONNECTION_DRIVERCLASS = "xplus.connection.driverClassName";

    public static String PERSISTENCE_CONNECTION_URL = "xplus.connection.url";

    public static String PERSISTENCE_CONNECTION_USER = "xplus.connection.user";

    public static String PERSISTENCE_CONNECTION_PASSWORD = "xplus.connection.password";

    public static String PERSISTENCE_CREATE_TABLE = "xplus.create-table";

    public static String PERSISTENCE_TRANSACTION_MANAGER = "xplus.transaction.manager";

    /**
     * persistence.xml file stream
     */
    private InputStream xmlFileStream;

    /**
     * persistence unit path
     */
    private String unitPath;

    /**
     * persistence units infomation
     */
    private List<PersistenceInfo> persistenceInfos = new ArrayList<PersistenceInfo>();

    private Logger logger = LogFactory.getLogger(PersistenceXMLParser.class);

    /**
     * default constructor
     */
    public PersistenceXMLParser() {
    }

    /**
     * constructor from a given file
     * @param xmlFile
     */
    public PersistenceXMLParser(File xmlFile) {
        try {
            this.xmlFileStream = xmlFile.toURL().openStream();
        } catch (MalformedURLException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * generate infomation
     */
    public void parse() {
        Document doc = loadFile();
        generateInfo(doc);
    }

    public List<PersistenceInfo> getPersistenceInfos() {
        return persistenceInfos;
    }

    public void setXmlFile(File xmlFile) {
        try {
            this.xmlFileStream = xmlFile.toURL().openStream();
        } catch (MalformedURLException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void setXmlFileStream(InputStream xmlFileStream) {
        this.xmlFileStream = xmlFileStream;
    }

    public void setUnitPath(String unitPath) {
        this.unitPath = unitPath;
    }

    /**
     * parse inputstram, build the xml document structure
     * @return xml document
     */
    private Document loadFile() {

        if (xmlFileStream == null)
            return null;

        Document doc = null;
        DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        docBuilderFactory.setValidating(false);
        InputSource source = new InputSource(xmlFileStream);
        DocumentBuilder docBuilder;
        try {
            docBuilder = docBuilderFactory.newDocumentBuilder();
            doc = docBuilder.parse(source);
        } catch (Exception pe) {
            UnitDeployingException ex = new UnitDeployingException(pe, UnitDeployingException.PERSISTENCE_XML_PARSING_ERROR);
            ex.showMessage();
        }

        return doc;
    }

    /**
     * generate persistence info
     * @param doc the persistence.xml document
     */
    private void generateInfo(Document doc) {
        Element top = doc.getDocumentElement();
        NodeList children = top.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            if (children.item(i).getNodeType() == Node.ELEMENT_NODE) {
                Element element = (Element) children.item(i);
                String tag = element.getTagName();
                if (tag.equals("persistence-unit")) {
                    PersistenceInfo info = buildPersistenceUnit(element);
                    persistenceInfos.add(info);
                }
            }
        }
    }

    /**
     * parse the elements in persistence.xml
     * @param unit xml document element
     * @return PersistenceInfo
     */
    private PersistenceInfo buildPersistenceUnit(Element unit) {
        PersistenceInfo info = new PersistenceInfo();
        info.setPersistenceUnitRootUrlPath(unitPath);

        String unitName = unit.getAttribute(PERSISTENCE_NAME);
        if (unitName != null) {
            logger.info("parsing persistence unit: " + unitName);
            info.setPersistenceUnitName(unitName);
        }
        info.setTransactionTypeName(unit.getAttribute(PERSISTENCE_TX_TYPE));

        NodeList children = unit.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            if (children.item(i).getNodeType() == Node.ELEMENT_NODE) {
                Element element = (Element) children.item(i);
                String tag = element.getTagName();
                if (tag.equals(PERSISTENCE_NON_JTA_DS)) {
                    Properties properties = new Properties();
                    properties.put(PERSISTENCE_DS_NAME, getElementContent(element));
                    JNDIDataSource jndiDataSource = new JNDIDataSource();
                    jndiDataSource.configure(properties);
                    info.setNonJtaDataSource(jndiDataSource);
                    info.addProperty(PersistenceXMLParser.PERSISTENCE_DB_NAME, properties
                            .getProperty(PersistenceXMLParser.PERSISTENCE_DB_NAME));

                } else if (tag.equals(PERSISTENCE_JTA_DS)) {
                    Properties properties = new Properties();
                    properties.put(PERSISTENCE_DS_NAME, getElementContent(element));
                    JNDIDataSource jndiDataSource = new JNDIDataSource();
                    jndiDataSource.configure(properties);
                    info.setJtaDataSource(jndiDataSource);
                    info.addProperty(PersistenceXMLParser.PERSISTENCE_DB_NAME, properties
                            .getProperty(PersistenceXMLParser.PERSISTENCE_DB_NAME));
                } else if (tag.equals(PERSISTENCE_PROVIDER)) {
                    String providerName = getElementContent(element);
                    if ((providerName == null) || providerName.equals(""))
                        providerName = "org.ow2.xplus.XplusPersistence";
                    info.setPersistenceProviderClassName(providerName);
                } else if (tag.equals(PERSISTENCE_CLASS)) {
                    info.addManagedClassName(getElementContent(element));
                } else if (tag.equals(PERSISTENCE_MAPPING_FILE)) {
                    info.addMappingFileName(getElementContent(element));
                } else if (tag.equals(PERSISTENCE_JAR_FILE)) {
                    info.addJarFileUrlPath(getElementContent(element));
                } else if (tag.equals(PERSISTENCE_EXCLUDE_UNLISTED)) {
                    info.setExcludeUnlistedClasses(true);
                } else if (tag.equals(PERSISTENCE_PROPERTIES)) {
                    NodeList props = element.getChildNodes();
                    for (int j = 0; j < props.getLength(); j++) {
                        if (props.item(j).getNodeType() == Node.ELEMENT_NODE) {
                            Element propElement = (Element) props.item(j);
                            if (propElement.getTagName().equals(PERSISTENCE_PROPERTY)) {
                                String propName = propElement.getAttribute(PERSISTENCE_PROPERTY_NAME);
                                String propValue = propElement.getAttribute(PERSISTENCE_PROPERTY_VALUE);
                                info.addProperty(propName, propValue);
                            }
                        }
                    }
                    buildDataSourceFromProperties(info);
                }
            }
        }

        return info;
    }

    /**
     * build DataSource from properties in PersistenceInfo
     * @param info
     */
    private void buildDataSourceFromProperties(PersistenceInfo info) {
        if ((info.getJtaDataSource() == null) && (info.getNonJtaDataSource() == null)) {
            StandaloneDataSource standaloneDataSource = new StandaloneDataSource();
            standaloneDataSource.configure(info.getProperties());
            info.setJtaDataSource(standaloneDataSource);
        }

    }

    /**
     * get xml Element Content
     * 
     * @param element xml Element
     * @return content
     */
    private String getElementContent(Element element) {

        NodeList children = element.getChildNodes();
        String result = null;
        for (int i = 0; i < children.getLength(); i++) {
            if (children.item(i).getNodeType() == Node.TEXT_NODE || children.item(i).getNodeType() == Node.CDATA_SECTION_NODE) {
                //result += children.item(i).getNodeValue();
                result = children.item(i).getNodeValue();
            }
        }
        return result;
    }

}
