/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.mapping;

import java.io.File;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import javax.persistence.Embeddable;
import javax.persistence.Entity;
import javax.persistence.MappedSuperclass;
import javax.persistence.spi.PersistenceUnitInfo;

import org.ow2.xplus.exception.MappingException;
import org.ow2.xplus.mapping.annotation.AnnotationProcessor;
import org.ow2.xplus.mapping.descriptor.MappingXMLProcessor;
import org.ow2.xplus.metadata.MetaDataRepository;
import org.ow2.xplus.metadata.UnitMetaData;

/**
 * An object-relational mapping parser allows the application to specify
 * annotations and mapping xml files to be used when creating a persistence
 * manager.
 * @author Lin Liang
 */

public class MappingProcessor {
    /**
     * persistence unit name
     */
    private String unitName;

    /**
     * the persistence Info of this unit
     */
    private PersistenceUnitInfo persistenceInfo;

    /**
     * constructor
     * @param persistenceInfo
     */
    public MappingProcessor(PersistenceUnitInfo persistenceInfo) {
        this.persistenceInfo = persistenceInfo;
        unitName = persistenceInfo.getPersistenceUnitName();
        UnitMetaData unitMetaData = MetaDataRepository.getUnitMetaData(unitName);
        unitMetaData.setDatabaseName(persistenceInfo.getProperties().getProperty("database-name"));
    }

    /**
     * process the unit, both annotation and mapping xml
     */
    public void process() {
        processAnnotation();
        processXML();
    }

    /**
     * process the or-mapping xml of this unit
     */
    private void processXML() {
        List<String> mappingFileNames = persistenceInfo.getMappingFileNames();

        if (mappingFileNames != null) {
            MappingXMLProcessor xmlProcessor = new MappingXMLProcessor();
            for (String mappingFileName : mappingFileNames) {
                File mappingFile = new File(mappingFileName);
                if ((mappingFile == null) || (!mappingFile.exists())) {
                    MappingException me = new MappingException("OR-mapping file not existed.",
                            MappingException.MAPIING_FILE_NOT_FOUND_ERROR);
                    me.showMessage();
                }
                xmlProcessor.process(mappingFile);
            }
        }
    }

    /**
     * process the annotations
     */
    private void processAnnotation() {
        List<String> managedClassNames = persistenceInfo.getManagedClassNames();

        List<Class> managedClasses = new ArrayList<Class>();

        if (managedClassNames != null) {
            for (String managedClassName : managedClassNames) {
                Class managedClass = null;
                try {
                    managedClass = Class.forName(managedClassName, true, Thread.currentThread().getContextClassLoader());
                } catch (ClassNotFoundException e) {
                    MappingException me = new MappingException(e, MappingException.CLASS_NOT_FOUND_ERROR);
                    me.showMessage();
                }
                if (managedClass != null)
                    managedClasses.add(managedClass);
            }
        }

        buildClassesHierarchy(managedClasses);
        AnnotationProcessor annProcessor = null;

        // process annotation
        for (Class managedClass : managedClasses) {
            annProcessor = new AnnotationProcessor(unitName, managedClass);
            annProcessor.process();
        }
        // build entity relation
        for (Class managedClass : managedClasses) {
            annProcessor = new AnnotationProcessor(unitName, managedClass);
            annProcessor.processRelation();
        }
        // build the inverse relation betweeen entity
        for (Class managedClass : managedClasses) {
            annProcessor = new AnnotationProcessor(unitName, managedClass);
            annProcessor.processInverseRelation();
        }
        // order all the classes to make embadded class process first
        List<Class> orderedEmbaddableClasses = new LinkedList<Class>();
        for (Class managedClass : managedClasses) {
            annProcessor = new AnnotationProcessor(unitName, managedClass);
            annProcessor.orderEmbaddable(orderedEmbaddableClasses);
        }
        List<Class> orderedClasses = new ArrayList<Class>();
        orderedClasses.addAll(orderedEmbaddableClasses);
        for (Class managedClass : managedClasses) {
            if (!orderedClasses.contains(managedClass))
                orderedClasses.add(managedClass);
        }
        // bind the embadded class to the metadata
        for (Class managedClass : orderedClasses) {
            annProcessor = new AnnotationProcessor(unitName, managedClass);
            annProcessor.bindEmbedded();
        }
        // build all the or-mapping metadata
        for (Class managedClass : orderedClasses) {
            annProcessor = new AnnotationProcessor(unitName, managedClass);
            annProcessor.buildMapping();
        }
    }

    /**
     * sort classes by its hierarchy, so the root class comes first
     */
    private void buildClassesHierarchy(List<Class> managedClasses) {
        List<Class> sortedClasses = new ArrayList<Class>();

        for (Class managedClass : managedClasses) {
            if (sortedClasses.contains(managedClass))
                continue;

            Stack<Class> hierarchyClasses = new Stack<Class>();
            getClassHierarchy(managedClass, hierarchyClasses);

            while (!hierarchyClasses.empty()) {
                Class hierarchyClass = hierarchyClasses.pop();
                if (!sortedClasses.contains(hierarchyClass)) {
                    sortedClasses.add(hierarchyClass);
                }
            }
        }

        managedClasses.clear();
        managedClasses.addAll(sortedClasses);
    }

    /**
     * get the inherience hierarchy of the entityClass
     * @param entityClass 
     * @param hierarchyClasses the hierarchy of the entity class
     */
    @SuppressWarnings("unchecked")
    private void getClassHierarchy(Class entityClass, Stack<Class> hierarchyClasses) {
        Class parent = entityClass;
        while ((parent != Object.class)) {
            if (parent.isAnnotationPresent(Entity.class) || parent.isAnnotationPresent(MappedSuperclass.class)
                    || parent.isAnnotationPresent(Embeddable.class)) {
                hierarchyClasses.push(parent);
            }
            parent = parent.getSuperclass();
        }
    }
}
