/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.mapping.annotation;

import java.lang.annotation.Annotation;
import java.lang.reflect.AnnotatedElement;
import java.lang.reflect.Field;
import java.util.List;

import javax.persistence.AttributeOverride;
import javax.persistence.AttributeOverrides;
import javax.persistence.Column;
import javax.persistence.DiscriminatorColumn;
import javax.persistence.DiscriminatorType;
import javax.persistence.DiscriminatorValue;
import javax.persistence.Embeddable;
import javax.persistence.Entity;
import javax.persistence.IdClass;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.MappedSuperclass;
import javax.persistence.PrimaryKeyJoinColumn;
import javax.persistence.PrimaryKeyJoinColumns;
import javax.persistence.SecondaryTable;
import javax.persistence.SecondaryTables;
import javax.persistence.SequenceGenerator;
import javax.persistence.Table;
import javax.persistence.TableGenerator;
import javax.persistence.UniqueConstraint;

import org.ow2.xplus.logger.LogFactory;
import org.ow2.xplus.logger.Logger;
import org.ow2.xplus.metadata.MetaDataRepository;
import org.ow2.xplus.metadata.UnitMetaData;
import org.ow2.xplus.metadata.object.ClassMetaData;
import org.ow2.xplus.metadata.relational.RdbColumn;
import org.ow2.xplus.metadata.relational.RdbDiscriminator;
import org.ow2.xplus.metadata.relational.RdbIdGenerator;
import org.ow2.xplus.metadata.relational.RdbInheritanceJoinedKey;
import org.ow2.xplus.metadata.relational.RdbPrimaryKey;
import org.ow2.xplus.metadata.relational.RdbTable;
import org.ow2.xplus.metadata.relational.RdbUniqueConstraint;
import org.ow2.xplus.metadata.strategy.DiscriminatorStrategy;
import org.ow2.xplus.metadata.strategy.InheritanceStrategy;
import org.ow2.xplus.metadata.strategy.discriminator.CharDiscriminatorStrategy;
import org.ow2.xplus.metadata.strategy.discriminator.IntegerDiscriminatorStrategy;
import org.ow2.xplus.metadata.strategy.discriminator.StringDiscriminatorStrategy;
import org.ow2.xplus.metadata.strategy.generator.IdentityGeneratorStrategy;
import org.ow2.xplus.metadata.strategy.generator.SequenceGeneratorStrategy;
import org.ow2.xplus.metadata.strategy.generator.TableGeneratorStrategy;
import org.ow2.xplus.metadata.strategy.inheritance.JoinedInheritanceStrategy;
import org.ow2.xplus.metadata.strategy.inheritance.SingleTableInheritanceStrategy;
import org.ow2.xplus.metadata.strategy.inheritance.TablePerClassInheritanceStrategy;
import org.ow2.xplus.util.NamingStrategy;

/**
 * annotation processor Will read the annotation from classes, apply the
 * principles of the EJB3 spec and produces the PKUAS or-mapping metadata
 * @author Lin Liang
 */

@SuppressWarnings("unchecked")
public class AnnotationProcessor {

    private UnitMetaData unitMetaData;

    private ClassMetaData classMetaData = null;

    private Class managedClass = null;

    private Logger logger = LogFactory.getLogger(AnnotationProcessor.class);

    private RdbTable primaryTable = null;

    /**
     * constructor
     * @param managedClass the entity class to be processed
     */
    public AnnotationProcessor(String unitName, Class managedClass) {
        this.managedClass = managedClass;
        unitMetaData = MetaDataRepository.getUnitMetaData(unitName);
        classMetaData = unitMetaData.getClassMetaData(managedClass.getName());
    }

    /**
     * process annotations in this class
     * @param managedClass
     */
    public void process() {
        if (!isPersistence(managedClass))
            return;

        logger.debug("processing " + managedClass.getName());

        parseORMapping();
        parseQueries();
        parseCallbacks();
    }

    /**
     * process entity relation many2many, many2one, one2one, one2many
     */
    public void processRelation() {
        if (!isPersistence(managedClass))
            return;

        Field[] fields = managedClass.getDeclaredFields();
        for (Field field : fields) {
            Member fieldMember = new Member(field);
            MemberAnnotationProcessor memberProcessor = new MemberAnnotationProcessor(unitMetaData.getUnitName(), classMetaData,
                    fieldMember);
            memberProcessor.processRelationship();
        }
    }

    /**
     * process the Inverse Relation from target to relation owner
     */
    public void processInverseRelation() {
        if (!isPersistence(managedClass))
            return;

        Field[] fields = managedClass.getDeclaredFields();
        for (Field field : fields) {
            Member fieldMember = new Member(field);
            MemberAnnotationProcessor memberProcessor = new MemberAnnotationProcessor(unitMetaData.getUnitName(), classMetaData,
                    fieldMember);
            memberProcessor.processInverseRelationship();
        }
    }

    /**
     * add all embedded classes to the list
     * @param orderedClasses
     */
    public void orderEmbaddable(List<Class> orderedClasses) {
        if (!managedClass.isAnnotationPresent(Embeddable.class) || orderedClasses.contains(managedClass)) {
            return;
        }

        orderedClasses.add(managedClass);
        int index = orderedClasses.indexOf(managedClass);

        Field[] fields = managedClass.getDeclaredFields();
        for (Field field : fields) {
            Member fieldMember = new Member(field);
            MemberAnnotationProcessor memberProcessor = new MemberAnnotationProcessor(unitMetaData.getUnitName(), classMetaData,
                    fieldMember);
            memberProcessor.orderEmbaddable(orderedClasses, index);
        }
    }

    /*
     * check whether the managedClass is a entity class
     */
    private boolean isPersistence(Class managedClass) {
        if (managedClass != null) {
            if ((managedClass.isAnnotationPresent(Entity.class)) || (managedClass.isAnnotationPresent(Embeddable.class))
                    || (managedClass.isAnnotationPresent(MappedSuperclass.class))) {
                return true;
            }
        }
        return false;
    }

    /**
     * parse life cycle callback
     */
    private void parseCallbacks() {
        logger.debug("parsing callback metadata...");
        CallbackProcessor callbackProcessor = new CallbackProcessor(unitMetaData);
        callbackProcessor.process(managedClass);
    }

    /**
     * parse the queries metadata described in Chapter 8 of JSR220
     */
    private void parseQueries() {
        logger.debug("parsing queries metadata...");
        QueryProcessor queryProcessor = new QueryProcessor(unitMetaData);
        queryProcessor.process(managedClass);
    }

    /**
     * parse the OR-mapping metadata described in Chapter 9 of JSR220
     */
    private void parseORMapping() {

        logger.debug("parsing or-mapping metadata...");

        if (managedClass.isAnnotationPresent(Entity.class)) {
            classMetaData.setHasTable(true);
        } else if (managedClass.isAnnotationPresent(Embeddable.class)) {
            classMetaData.setEmbeded(true);
            classMetaData.setHasTable(false);
        } else if (managedClass.isAnnotationPresent(MappedSuperclass.class)) {
            classMetaData.setMappedSuperclass(true);
        } else
            return;

        processInheritance();

        processTable();

        prepareDefaultGenerator();
        processSequenceGenerator();
        processTableGenerator();

        processAttributeOverrides();

        processIdClass();

        if (classMetaData.isInheritanceRoot()) {
            if (classMetaData.getInheritanceType() == InheritanceType.JOINED
                    || classMetaData.getInheritanceType() == InheritanceType.SINGLE_TABLE) {
                processDiscriminator();
            }
        }

        processDiscriminatorValue();

        if (classMetaData.getInheritanceType() == InheritanceType.JOINED) {
            processPrimaryKeyJoinColumns();
        }

        MemberAnnotationProcessor memberProcessor = null;

        Field[] fields = managedClass.getDeclaredFields();
        for (Field field : fields) {
            Member fieldMember = new Member(field);
            memberProcessor = new MemberAnnotationProcessor(unitMetaData.getUnitName(), classMetaData, fieldMember);
            memberProcessor.process();
        }

        if (!classMetaData.isInheritanceRoot())
            bindPrimaryKey();

        processSecondaryTables();
    }

    /**
     * bind the primary key columns to table
     */
    private void bindPrimaryKey() {
        Class root = classMetaData.getInheritanceRootClass();
        ClassMetaData rootMetaData = unitMetaData.getClassMetaData(root.getName());
        RdbTable rootTable = unitMetaData.getTable(rootMetaData.getPrimaryTableName());
        RdbPrimaryKey primaryKey = rootTable.getPrimaryKey();
        primaryTable.setPrimaryKey(primaryKey);
    }

    /**
     * process
     * @SecondaryTables annotation
     */
    private void processSecondaryTables() {
        SecondaryTables secondaryTablesAnn = getAnnotation(managedClass, SecondaryTables.class);
        if (secondaryTablesAnn != null) {
            SecondaryTable[] secondaryTableAnns = secondaryTablesAnn.value();
            for (SecondaryTable secondaryTableAnn : secondaryTableAnns) {
                processSecondaryTable(secondaryTableAnn);
            }
        }

        SecondaryTable secondaryTableAnn = getAnnotation(managedClass, SecondaryTable.class);
        if (secondaryTableAnn != null) {
            processSecondaryTable(secondaryTableAnn);
        }
    }

    /**
     * process
     * @SecondaryTable annotation The SecondaryTable annotation is used to
     *                 specify a secondary table for the annotated entity class.
     * @param secondaryTableAnn
     */
    private void processSecondaryTable(SecondaryTable secondaryTableAnn) {
        String tableName = NamingStrategy.getFullTableName(secondaryTableAnn.catalog(), secondaryTableAnn.schema(),
                secondaryTableAnn.name());

        RdbTable secondaryTable = unitMetaData.getTable(tableName);

        UniqueConstraint[] uniqueConstraints = secondaryTableAnn.uniqueConstraints();

        for (UniqueConstraint uniqueConstraint : uniqueConstraints) {
            String[] columnNames = uniqueConstraint.columnNames();
            RdbUniqueConstraint rdbUniqueConstraint = new RdbUniqueConstraint(columnNames);
            secondaryTable.AddUniqueConstraint(rdbUniqueConstraint);
        }

        PrimaryKeyJoinColumn[] primaryKeyJoinColumnAnns = secondaryTableAnn.pkJoinColumns();
        for (PrimaryKeyJoinColumn primaryKeyJoinColumnAnn : primaryKeyJoinColumnAnns) {
            processSecondaryTablePrimaryKeyJoinColumn(primaryKeyJoinColumnAnn, secondaryTable);
        }

        // if no annotation of @PrimaryKeyJoinColumn, use primary key as default
        // if (primaryKeyJoinColumnAnns.length == 0){
        RdbPrimaryKey primaryKey = primaryTable.getPrimaryKey();
        for (RdbColumn pkColumn : primaryKey.getColumns()) {
            RdbColumn column = new RdbColumn(pkColumn);
            column.setPrimaryKey(true);
            column.setTableName(secondaryTable.getCanonicalName());
            secondaryTable.addColumn(column);
            secondaryTable.addAttributeOverride(column.getName(), column);
        }
        // }

        primaryTable.addSecondaryTableName(tableName);
    }

    /**
     * process SecondaryTable
     * @PrimaryKeyJoinColumn annotation it is used with
     * @SecondaryTable annotation to join a secondary table to a primary table
     * @param primaryKeyJoinColumnAnn
     * @param secondaryTable
     */
    private void processSecondaryTablePrimaryKeyJoinColumn(PrimaryKeyJoinColumn primaryKeyJoinColumnAnn, RdbTable secondaryTable) {
        RdbColumn override = new RdbColumn(primaryKeyJoinColumnAnn.name());
        override.setColumnDefinition(primaryKeyJoinColumnAnn.columnDefinition());
        secondaryTable.addAttributeOverride(primaryKeyJoinColumnAnn.referencedColumnName().toUpperCase(), override);
    }

    /**
     * process
     * @PrimaryKeyJoinColumns annotation Composite foreign keys are supported by
     *                        means of the PrimaryKeyJoinColumns annotation.
     */
    private void processPrimaryKeyJoinColumns() {
        RdbInheritanceJoinedKey inheritanceJoinedKey = new RdbInheritanceJoinedKey();
        // process @PrimaryKeyJoinColumns
        PrimaryKeyJoinColumns primaryKeyJoinColumnsAnn = getAnnotation(managedClass, PrimaryKeyJoinColumns.class);
        if (primaryKeyJoinColumnsAnn != null) {
            PrimaryKeyJoinColumn[] primaryKeyJoinColumnAnns = primaryKeyJoinColumnsAnn.value();
            for (PrimaryKeyJoinColumn primaryKeyJoinColumnAnn : primaryKeyJoinColumnAnns) {
                processPrimaryKeyJoinColumn(primaryKeyJoinColumnAnn, inheritanceJoinedKey);
            }
        }
        // process @PrimaryKeyJoinColumn
        PrimaryKeyJoinColumn primaryKeyJoinColumnAnn = getAnnotation(managedClass, PrimaryKeyJoinColumn.class);
        if (primaryKeyJoinColumnAnn != null) {
            processPrimaryKeyJoinColumn(primaryKeyJoinColumnAnn, inheritanceJoinedKey);
        }

        primaryTable.setInheritanceJoinedKey(inheritanceJoinedKey);
    }

    /**
     * process
     * @PrimaryKeyJoinColumn annotation The PrimaryKeyJoinColumn annotation is
     *                       used to join the primary table of an entity
     *                       subclass in the JOINED mapping strategy to the
     *                       primary table of its superclass; and it may be used
     *                       in a OneToOne mapping in which the primary key of
     *                       the referencing entity is used as a foreign key to
     *                       the referenced entity.
     */
    private void processPrimaryKeyJoinColumn(PrimaryKeyJoinColumn primaryKeyJoinColumnAnn,
            RdbInheritanceJoinedKey inheritanceJoinedKey) {
        // get the inheritance parent
        String inheritanceEntityName = classMetaData.getParent().getName();
        inheritanceJoinedKey.setInheritanceEntityName(inheritanceEntityName);
        String inheritanceEntityTableName = unitMetaData.getClassMetaData(inheritanceEntityName).getPrimaryTableName();
        inheritanceJoinedKey.setInheritanceEntityTableName(inheritanceEntityTableName);

        RdbTable inheritanceEntityTable = unitMetaData.getTable(inheritanceEntityTableName);
        RdbPrimaryKey inheritancePrimaryKey = inheritanceEntityTable.getPrimaryKey();
        // bind inheritance PrimaryKey columns
        if (!inheritancePrimaryKey.getColumns().isEmpty()) {
            String inheritancePrimaryColumnName = primaryKeyJoinColumnAnn.referencedColumnName().equals("") ? inheritancePrimaryKey
                    .getColumns().get(0).getName()
                    : primaryKeyJoinColumnAnn.referencedColumnName();

            RdbColumn inheritancePrimaryColumn = inheritancePrimaryKey.getColumn(inheritancePrimaryColumnName);
            String joinColumnName = primaryKeyJoinColumnAnn.name().equals("") ? inheritancePrimaryKey.getColumns().get(0)
                    .getName() : primaryKeyJoinColumnAnn.name();
            RdbColumn joinColumn = new RdbColumn(inheritancePrimaryColumn);
            joinColumn.setPrimaryKey(true);
            joinColumn.setName(joinColumnName);
            joinColumn.setColumnDefinition(primaryKeyJoinColumnAnn.columnDefinition());
            joinColumn.setTableName(primaryTable.getCanonicalName());
            primaryTable.addColumn(joinColumn);

            inheritanceJoinedKey.addcolumnMaping(joinColumnName, inheritancePrimaryColumnName);
        } else {
            // will bind it later
            inheritanceJoinedKey.addcolumnMaping(primaryKeyJoinColumnAnn.name(), primaryKeyJoinColumnAnn.referencedColumnName());
        }
    }

    /**
     * process
     * @Discriminator annotation The DiscriminatorColumn annotation is used to
     *                define the discriminator column for the SINGLE_TABLE and
     *                JOINED inheritance mapping strategies.
     */
    private void processDiscriminator() {
        DiscriminatorColumn discriminatorColumnAnn = getAnnotation(managedClass, DiscriminatorColumn.class);
        if (discriminatorColumnAnn == null)
            return;
        // build column
        RdbDiscriminator discriminator = new RdbDiscriminator();
        String discriminatorColumnName = discriminatorColumnAnn.name().equals("") ? "DType" : discriminatorColumnAnn.name();

        RdbColumn discriminatorColumn = new RdbColumn(discriminatorColumnName);
        discriminatorColumn.setColumnDefinition(discriminatorColumnAnn.columnDefinition());
        discriminatorColumn.setTableName(primaryTable.getCanonicalName());
        discriminator.setDiscriminatorColumnName(discriminatorColumnName);

        // process @DiscriminatorType
        DiscriminatorType discriminatorType = discriminatorColumnAnn.discriminatorType();
        DiscriminatorStrategy discriminatorStrategy = null;
        if (discriminatorType == DiscriminatorType.CHAR) {
            discriminatorColumn.setJavaType(Character.class);
            discriminatorStrategy = new CharDiscriminatorStrategy();
        } else if (discriminatorType == DiscriminatorType.INTEGER) {
            discriminatorColumn.setJavaType(Integer.class);
            discriminatorStrategy = new IntegerDiscriminatorStrategy();
        } else if (discriminatorType == DiscriminatorType.STRING) {
            discriminatorColumn.setJavaType(String.class);
            discriminatorStrategy = new StringDiscriminatorStrategy(discriminatorColumnAnn.length());
        }
        discriminator.setDiscriminatorStrategy(discriminatorStrategy);
        discriminatorColumn.setSqlType(discriminatorStrategy.getSqlType());

        // process @DiscriminatorValue
        DiscriminatorValue discriminatorValueAnn = getAnnotation(managedClass, DiscriminatorValue.class);
        if (!(discriminatorValueAnn == null)) {
            String discriminatorValue = discriminatorValueAnn.value();
            if (discriminatorValue == null) {
                discriminatorValue = (String) discriminatorStrategy.getValue(managedClass.getName());
            }
            classMetaData.setDiscriminatorValue(discriminatorValue);
        }

        primaryTable.setDiscriminator(discriminator);
        primaryTable.addColumn(discriminatorColumn);
    }

    /**
     * process
     * @DiscriminatorValue annotation
     */
    private void processDiscriminatorValue() {
        DiscriminatorValue discriminatorValueAnn = getAnnotation(managedClass, DiscriminatorValue.class);
        if (!(discriminatorValueAnn == null)) {
            String discriminatorValue = discriminatorValueAnn.value();
            if (discriminatorValue != null) {
                classMetaData.setDiscriminatorValue(discriminatorValue);
            }
        }
    }

    /**
     * process
     * @IdClass annotation The IdClass annotation is applied to an entity class
     *          or a mapped superclass to specify a composite primary key class
     *          that is mapped to multiple fields or properties of the entity.
     */
    private void processIdClass() {
        IdClass idClassAnn = getAnnotation(managedClass, IdClass.class);
        if (idClassAnn == null)
            return;

        Class idClass = idClassAnn.value();
        RdbPrimaryKey primaryKey = new RdbPrimaryKey();
        primaryKey.setPkClassName(idClass.getName());
        primaryTable.setPrimaryKey(primaryKey);
    }

    /**
     * bind the id class columns to the primary table
     */
    private void bindIdClass() {
        String pkClassName = primaryTable.getPrimaryKey().getPkClassName();
        if (pkClassName == null)
            return;
        RdbPrimaryKey primaryKey = primaryTable.getPrimaryKey();
        ClassMetaData idClassMetaData = unitMetaData.getClassMetaData(pkClassName);
        String idClassTableName = idClassMetaData.getPrimaryTableName();
        RdbTable idClassTable = unitMetaData.getTable(idClassTableName);

        // bind columns from idClass Table
        List<RdbColumn> columns = idClassTable.getColumns();
        for (RdbColumn column : columns) {
            if (!primaryKey.getColumns().contains(column)) {
                column.setPrimaryKey(true);
                primaryKey.addColumn(column);
            }

            String fieldName = NamingStrategy.unqualify(column.getCanonicalFieldName());
            try {
                classMetaData.getFieldMetaData(fieldName).getValue().getSimpleValue().getColumns().get(0).setName(
                        column.getName());
                primaryTable.getColumn(fieldName).setName(column.getName());
            } catch (NullPointerException ne) {
                // TODO
                ne.printStackTrace();
            }
        }
    }

    /**
     * defines the default primary key generator
     */
    private void prepareDefaultGenerator() {
        // TODO prepare default id generator

        TableGeneratorStrategy tableGeneratorStrategy = new TableGeneratorStrategy(unitMetaData);
        RdbIdGenerator tableIdGen = new RdbIdGenerator(UnitMetaData.DEFAULT_TABLE_GENERATOR_NAME);
        tableIdGen.setGeneratorStrategy(tableGeneratorStrategy);
        unitMetaData.addTableIdGen(primaryTable.getName(), tableIdGen);

        // SequenceGeneratorStrategy sequenceGeneratorStrategy =
        // new SequenceGeneratorStrategy();
        IdentityGeneratorStrategy identityGeneratorStrategy = new IdentityGeneratorStrategy();

        RdbIdGenerator sequenceIdGen = new RdbIdGenerator(UnitMetaData.DEFAULT_SEQUENCE_GENERATOR_NAME);
        sequenceIdGen.setGeneratorStrategy(identityGeneratorStrategy);
        unitMetaData.addTableIdGen(primaryTable.getName(), sequenceIdGen);
    }

    /**
     * bind embedded class columns to primary table
     */
    public void bindEmbedded() {
        if (!isPersistence(managedClass))
            return;

        primaryTable = unitMetaData.getTable(classMetaData.getPrimaryTableName());

        if (managedClass.isAnnotationPresent(IdClass.class))
            bindIdClass();

        MemberAnnotationProcessor memberProcessor = null;

        Field[] fields = managedClass.getDeclaredFields();
        for (Field field : fields) {
            Member fieldMember = new Member(field);
            memberProcessor = new MemberAnnotationProcessor(unitMetaData.getUnitName(), classMetaData, fieldMember);
            memberProcessor.buildEmbeddedMapping();
        }
    }

    /**
     * build or-mapping bind AttributeOverrides, build primary table
     */
    public void buildMapping() {
        if (!isPersistence(managedClass))
            return;

        primaryTable = unitMetaData.getTable(classMetaData.getPrimaryTableName());
        if (managedClass.isAnnotationPresent(IdClass.class))
            bindIdClass();
        MemberAnnotationProcessor memberProcessor = null;

        // bind AttributeOverrides
        Field[] fields = managedClass.getDeclaredFields();
        for (Field field : fields) {
            Member fieldMember = new Member(field);
            memberProcessor = new MemberAnnotationProcessor(unitMetaData.getUnitName(), classMetaData, fieldMember);
            memberProcessor.buildRelationEmbeddedMapping();
            memberProcessor.bindAttributeOverride();
        }

        // build primary table
        primaryTable = unitMetaData.getTable(classMetaData.getPrimaryTableName());
        for (String secondaryTableName : primaryTable.getSecondaryTableNames()) {
            RdbTable secondaryTable = unitMetaData.getTable(secondaryTableName);
            RdbPrimaryKey primaryKey = primaryTable.getPrimaryKey();
            for (RdbColumn primaryColumn : primaryKey.getColumns()) {
                RdbColumn column = new RdbColumn(primaryColumn);
                RdbColumn overrideColumnInfo = secondaryTable.getAttributeOverride(primaryColumn.getName());
                if (overrideColumnInfo != null) {
                    column.setName(overrideColumnInfo.getName());
                    column.setColumnDefinition(overrideColumnInfo.getColumnDefinition());
                }
                column.setTableName(secondaryTable.getCanonicalName());
                secondaryTable.addColumn(column);
            }
        }

        InheritanceStrategy inheritanceStrategy = classMetaData.getInheritanceStrategy();
        inheritanceStrategy.buildTable(classMetaData.getUnitName(), managedClass.getName());
    }

    /**
     * process
     * @Inheritance The Inheritance annotation defines the inheritance strategy
     *              to be used for an entity class hierarchy.
     */
    private void processInheritance() {
        Inheritance inheritanceAnn = getAnnotation(managedClass, Inheritance.class);

        InheritanceType inheritanceType = null;
        // It is specified on the entity class that is the root of the entity
        // class hierarchy.
        if (classMetaData.isInheritanceRoot() && inheritanceAnn != null) {
            inheritanceType = inheritanceAnn.strategy();
        } else {
            Class root = classMetaData.getInheritanceRootClass();
            ClassMetaData rootMetaData = unitMetaData.getClassMetaData(root.getName());
            inheritanceType = rootMetaData.getInheritanceType();
        }
        // default type
        if (inheritanceType == null)
            inheritanceType = InheritanceType.SINGLE_TABLE;

        classMetaData.setInheritanceType(inheritanceType);

        InheritanceStrategy inheritanceStrategy = null;
        if (inheritanceType == InheritanceType.JOINED) {
            inheritanceStrategy = JoinedInheritanceStrategy.getInstance();
        } else if (inheritanceType == InheritanceType.TABLE_PER_CLASS) {
            inheritanceStrategy = TablePerClassInheritanceStrategy.getInstance();
        } else if (inheritanceType == InheritanceType.SINGLE_TABLE) {
            inheritanceStrategy = SingleTableInheritanceStrategy.getInstance();
        }
        classMetaData.setInheritanceStrategy(inheritanceStrategy);

    }

    /**
     * process
     * @Table annotation The Table annotation specifies the primary table for
     *        the annotated entity.
     */
    private void processTable() {

        Table tableAnn = getAnnotation(managedClass, Table.class);

        InheritanceStrategy inheritanceStrategy = classMetaData.getInheritanceStrategy();

        if (tableAnn != null) {
            String tableFullName = null;

            if (!inheritanceStrategy.ignoreTableAnn(classMetaData.getUnitName(), managedClass.getName())) {
                tableFullName = NamingStrategy.getFullTableName(tableAnn.catalog(), tableAnn.schema(), tableAnn.name());
            }

            if (tableFullName == null) {
                tableFullName = inheritanceStrategy.getTableName(classMetaData.getUnitName(), managedClass.getName());
            }

            RdbTable table = unitMetaData.getTable(tableFullName);

            if (!inheritanceStrategy.ignoreTableAnn(classMetaData.getUnitName(), managedClass.getName())) {
                if (!tableAnn.catalog().equals(""))
                    table.setCatalog(tableAnn.catalog());
                if (!tableAnn.schema().equals(""))
                    table.setSchema(tableAnn.schema());
                if (!tableAnn.name().equals(""))
                    table.setName(tableAnn.name());

                UniqueConstraint[] uniqueConstraints = tableAnn.uniqueConstraints();

                for (UniqueConstraint uniqueConstraint : uniqueConstraints) {
                    String[] columnNames = uniqueConstraint.columnNames();
                    RdbUniqueConstraint rdbUniqueConstraint = new RdbUniqueConstraint(columnNames);
                    table.AddUniqueConstraint(rdbUniqueConstraint);
                }
            }

            classMetaData.setPrimaryTableName(table.getCanonicalName());
        } else {
            classMetaData.setPrimaryTableName(inheritanceStrategy.getTableName(classMetaData.getUnitName(), managedClass
                    .getName()));
        }

        primaryTable = unitMetaData.getTable(classMetaData.getPrimaryTableName());

    }

    /**
     * process
     * @AttributeOverrides
     */
    private void processAttributeOverrides() {
        // process AttributeOverrides
        AttributeOverrides attributeOverridesAnn = getAnnotation(managedClass, AttributeOverrides.class);
        if (attributeOverridesAnn != null) {
            AttributeOverride[] attributeOverrideAnns = attributeOverridesAnn.value();
            for (AttributeOverride attributeOverrideAnn : attributeOverrideAnns) {
                processAttributeOverride(attributeOverrideAnn);
            }
        }
        // process AttributeOverride
        AttributeOverride attributeOverrideAnn = getAnnotation(managedClass, AttributeOverride.class);
        if (attributeOverrideAnn != null) {
            processAttributeOverride(attributeOverrideAnn);
        }
    }

    /**
     * process
     * @AttributeOverride
     * @param attributeOverrideAnn
     */
    private void processAttributeOverride(AttributeOverride attributeOverrideAnn) {
        String fieldName = attributeOverrideAnn.name();
        Column columnAnn = attributeOverrideAnn.column();

        RdbColumn column = new RdbColumn(columnAnn.name());
        column.setUnique(columnAnn.unique());
        column.setNullable(columnAnn.nullable());
        column.setInsertable(columnAnn.insertable());
        column.setUpdatable(columnAnn.updatable());
        column.setColumnDefinition(columnAnn.columnDefinition());
        column.setLength(new Integer(columnAnn.length()));
        column.setPrecision(new Integer(columnAnn.precision()));
        column.setScale(new Integer(columnAnn.scale()));
        column.setTableName(columnAnn.table());

        primaryTable.addAttributeOverride(fieldName, column);
    }

    private <T extends Annotation> T getAnnotation(AnnotatedElement annElement, Class<T> annotationClass) {
        T ann = null;
        if (annElement.isAnnotationPresent(annotationClass)) {
            ann = annElement.getAnnotation(annotationClass);
        }
        return ann;
    }

    /**
     * process @TableGenerator annotation
     * The TableGenerator annotation defines a primary key generator that may be referenced 
     * by name when a generator element is specified for the GeneratedValue annotation.
     *
     */
    private void processTableGenerator() {
        TableGenerator tableGeneratorAnn = getAnnotation(managedClass, TableGenerator.class);
        if (tableGeneratorAnn == null)
            return;

        TableGeneratorStrategy tableGeneratorStrategy = new TableGeneratorStrategy(unitMetaData);

        String tableFullName = NamingStrategy.getFullTableName(tableGeneratorAnn.catalog(), tableGeneratorAnn.schema(),
                tableGeneratorAnn.name());

        RdbTable table = unitMetaData.getTable(tableFullName);

        if (!tableGeneratorAnn.catalog().equals(""))
            table.setCatalog(tableGeneratorAnn.catalog());
        if (!tableGeneratorAnn.schema().equals(""))
            table.setSchema(tableGeneratorAnn.schema());
        if (!tableGeneratorAnn.name().equals(""))
            table.setName(tableGeneratorAnn.name());

        UniqueConstraint[] uniqueConstraints = tableGeneratorAnn.uniqueConstraints();

        for (UniqueConstraint uniqueConstraint : uniqueConstraints) {
            String[] columnNames = uniqueConstraint.columnNames();
            RdbUniqueConstraint rdbUniqueConstraint = new RdbUniqueConstraint(columnNames);
            table.AddUniqueConstraint(rdbUniqueConstraint);
        }

        tableGeneratorStrategy.setTableName(tableFullName);
        tableGeneratorStrategy.setPkColumnName(tableGeneratorAnn.pkColumnName());
        tableGeneratorStrategy.setValueColumnName(tableGeneratorAnn.valueColumnName());
        tableGeneratorStrategy.setPkColumnValue(tableGeneratorAnn.pkColumnValue());
        tableGeneratorStrategy.setInitialValue(tableGeneratorAnn.initialValue());
        tableGeneratorStrategy.setAllocationSize(tableGeneratorAnn.allocationSize());

        RdbIdGenerator idgen = new RdbIdGenerator();
        idgen.setGeneratorStrategy(tableGeneratorStrategy);
        unitMetaData.addTableIdGen(primaryTable.getName(), idgen);

        RdbColumn pkColumn = new RdbColumn(tableGeneratorStrategy.getPkColumnName());
        RdbColumn valueColumn = new RdbColumn(tableGeneratorStrategy.getValueColumnName());

        table.addColumn(pkColumn);
        table.addColumn(valueColumn);
        pkColumn.setTableName(tableFullName);
        valueColumn.setTableName(tableFullName);
    }

    /**
     * process @SequenceGenerator annotation
     * The SequenceGenerator annotation defines a primary key generator that may be referenced by
     * name when a generator element is specified for the GeneratedValue annotation.
     */
    private void processSequenceGenerator() {
        SequenceGenerator sequenceGeneratorAnn = getAnnotation(managedClass, SequenceGenerator.class);
        if (sequenceGeneratorAnn == null)
            return;

        SequenceGeneratorStrategy sequenceGeneratorStrategy = new SequenceGeneratorStrategy();

        sequenceGeneratorStrategy.setName(sequenceGeneratorAnn.name());
        sequenceGeneratorStrategy.setSequenceName(sequenceGeneratorAnn.sequenceName());
        sequenceGeneratorStrategy.setInitialValue(sequenceGeneratorAnn.initialValue());
        sequenceGeneratorStrategy.setAllocationSize(sequenceGeneratorAnn.allocationSize());

        RdbIdGenerator idgen = new RdbIdGenerator();
        idgen.setGeneratorStrategy(sequenceGeneratorStrategy);
        unitMetaData.addTableIdGen(primaryTable.getName(), idgen);
    }

}
