/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.mapping.annotation;

import java.lang.annotation.Annotation;
import java.lang.reflect.AnnotatedElement;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.ow2.xplus.util.ClassStrategy;
import org.ow2.xplus.util.NamingStrategy;

/**
 * Class member accessor
 * @author Lin Liang
 */

public class Member implements AnnotatedElement {

    private static String GET_PREFIX = "get";

    private static String SET_PREFIX = "set";

    private static String IS_PREFIX = "is";

    private Method getMethod = null;

    private Method setMethod = null;

    private Field field = null;

    private String propertyName;

    /**
     * a Class object that identifies the type class of the ParameterizedType
     * member
     */
    private Class parameterizedType;

    /**
     * a Class object that identifies the declared type for the field
     */
    private Class declaredType;

    private boolean isCollection = false;

    public Member(Field field) {
        this.field = field;
        propertyName = this.field.getName();

        String getMethodName = NamingStrategy.getFieldAccessMethod(field.getName(), GET_PREFIX);
        String isMethodName = NamingStrategy.getFieldAccessMethod(field.getName(), IS_PREFIX);
        String setMethodName = NamingStrategy.getFieldAccessMethod(field.getName(), SET_PREFIX);

        Class declaringClass = field.getDeclaringClass();
        for (Method method : declaringClass.getDeclaredMethods()) {
            if (method.getName().equals(getMethodName)) {
                getMethod = method;
            } else if (method.getName().equals(setMethodName)) {
                setMethod = method;
            } else if (method.getName().equals(isMethodName)) {
                getMethod = method;
            }
        }

        declaredType = field.getType();

        if (!checkCollection()) {
            parameterizedType = field.getType();
        } else {
            isCollection = true;
            parameterizedType = getCollectionGeneric();
        }

    }

    /**
     * get the raw type from Generic Type e.g. List<User> return List
     * @return the raw type
     */
    private Class getRawTypeFromGeneric() {
        if (field.getGenericType() instanceof ParameterizedType)
            return (Class) (((ParameterizedType) field.getGenericType()).getRawType());
        return field.getType();
    }

    /**
     * get the type class of a ParameterizedType from a generic collection e.g.
     * List<User> return User
     * @return the type class
     */
    private Class getCollectionGeneric() {
        ParameterizedType pType = null;

        Type type = field.getGenericType();

        if (type instanceof ParameterizedType) {
            pType = (ParameterizedType) type;
        } else {
            type = getMethod.getGenericReturnType();
            if (type instanceof ParameterizedType) {
                pType = (ParameterizedType) type;
            } else {
                type = setMethod.getGenericParameterTypes()[0];
                if (type instanceof ParameterizedType) {
                    pType = (ParameterizedType) type;
                }
            }
        }

        if (Map.class.isAssignableFrom(getRawTypeFromGeneric())) {
            return (Class) pType.getActualTypeArguments()[1];
        }
        return (Class) pType.getActualTypeArguments()[0];
    }

    /**
     * if the class is a Collection
     * @return true if it is a Collection type
     */
    private boolean checkCollection() {
        if (ClassStrategy.isLobType(getRawTypeFromGeneric()))
            return false;

        return Collection.class.isAssignableFrom(getRawTypeFromGeneric()) || Map.class.isAssignableFrom(getRawTypeFromGeneric())
                || getRawTypeFromGeneric().isArray();
    }

    /**
     * @return the name of the field
     */
    public String getPropertyName() {
        return propertyName;
    }

    /**
     * @return a Class object that identifies the type class of the
     *         ParameterizedType member
     */
    public Class getParameterizedType() {
        return parameterizedType;
    }

    public boolean isAnnotationPresent(Class<? extends Annotation> annotationType) {
        if (field.isAnnotationPresent(annotationType)) {
            return true;
        } else if (getMethod != null && getMethod.isAnnotationPresent(annotationType)) {
            return true;
        } else if (setMethod != null && setMethod.isAnnotationPresent(annotationType)) {
            return true;
        }
        return false;
    }

    public <T extends Annotation> T getAnnotation(Class<T> annotationType) {
        if (field.getAnnotation(annotationType) != null) {
            return field.getAnnotation(annotationType);
        } else if (getMethod != null && getMethod.getAnnotation(annotationType) != null) {
            return getMethod.getAnnotation(annotationType);
        } else if (setMethod != null) {
            return setMethod.getAnnotation(annotationType);
        }
        return null;
    }

    public Annotation[] getAnnotations() {
        List<Annotation> anns = new ArrayList<Annotation>();
        for (Annotation ann : field.getAnnotations()) {
            anns.add(ann);
        }
        if (getMethod != null) {
            for (Annotation ann : getMethod.getAnnotations()) {
                anns.add(ann);
            }
        }
        if (setMethod != null) {
            for (Annotation ann : setMethod.getAnnotations()) {
                anns.add(ann);
            }
        }
        return (Annotation[]) anns.toArray();
    }

    public Annotation[] getDeclaredAnnotations() {
        List<Annotation> anns = new ArrayList<Annotation>();
        for (Annotation ann : field.getDeclaredAnnotations()) {
            anns.add(ann);
        }
        if (getMethod != null) {
            for (Annotation ann : getMethod.getDeclaredAnnotations()) {
                anns.add(ann);
            }
        }
        if (setMethod != null) {
            for (Annotation ann : setMethod.getDeclaredAnnotations()) {
                anns.add(ann);
            }
        }
        return (Annotation[]) anns.toArray();
    }

    /**
     * @return a Class object that identifies the declared type for the field
     */
    public Class getDeclaredType() {
        return declaredType;
    }

    /**
     * @return true if this field is a Collection
     */
    public boolean isCollection() {
        return isCollection;
    }

    /**
     * @return true if this field is a Enum type
     */
    public boolean isEnum() {
        return declaredType.isEnum();
    }
}
