/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.mapping.annotation;

import java.lang.annotation.Annotation;
import java.lang.reflect.AnnotatedElement;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.persistence.AttributeOverride;
import javax.persistence.AttributeOverrides;
import javax.persistence.Basic;
import javax.persistence.Column;
import javax.persistence.Embeddable;
import javax.persistence.Embedded;
import javax.persistence.EmbeddedId;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.JoinColumns;
import javax.persistence.JoinTable;
import javax.persistence.Lob;
import javax.persistence.ManyToMany;
import javax.persistence.ManyToOne;
import javax.persistence.MapKey;
import javax.persistence.MappedSuperclass;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.OrderBy;
import javax.persistence.PrimaryKeyJoinColumn;
import javax.persistence.PrimaryKeyJoinColumns;
import javax.persistence.SequenceGenerator;
import javax.persistence.TableGenerator;
import javax.persistence.Temporal;
import javax.persistence.Transient;
import javax.persistence.UniqueConstraint;
import javax.persistence.Version;

import org.ow2.xplus.exception.MappingException;
import org.ow2.xplus.metadata.MetaDataRepository;
import org.ow2.xplus.metadata.UnitMetaData;
import org.ow2.xplus.metadata.object.ClassMetaData;
import org.ow2.xplus.metadata.object.CollectionValue;
import org.ow2.xplus.metadata.object.FieldMetaData;
import org.ow2.xplus.metadata.object.ListValue;
import org.ow2.xplus.metadata.object.ManyToManyRelation;
import org.ow2.xplus.metadata.object.ManyToOneRelation;
import org.ow2.xplus.metadata.object.MapValue;
import org.ow2.xplus.metadata.object.OneToManyRelation;
import org.ow2.xplus.metadata.object.OneToOneRelation;
import org.ow2.xplus.metadata.object.Relationship;
import org.ow2.xplus.metadata.object.SetValue;
import org.ow2.xplus.metadata.object.SimpleValue;
import org.ow2.xplus.metadata.object.ValueMetaData;
import org.ow2.xplus.metadata.relational.RdbColumn;
import org.ow2.xplus.metadata.relational.RdbForeignKey;
import org.ow2.xplus.metadata.relational.RdbIdGenerator;
import org.ow2.xplus.metadata.relational.RdbPrimaryKey;
import org.ow2.xplus.metadata.relational.RdbRelation;
import org.ow2.xplus.metadata.relational.RdbRelationTable;
import org.ow2.xplus.metadata.relational.RdbTable;
import org.ow2.xplus.metadata.relational.RdbUniqueConstraint;
import org.ow2.xplus.metadata.strategy.VersionStrategy;
import org.ow2.xplus.metadata.strategy.generator.IdentityGeneratorStrategy;
import org.ow2.xplus.metadata.strategy.generator.SequenceGeneratorStrategy;
import org.ow2.xplus.metadata.strategy.generator.TableGeneratorStrategy;
import org.ow2.xplus.metadata.strategy.value.BasicValueStrategy;
import org.ow2.xplus.metadata.strategy.value.EnumeratedValueStrategy;
import org.ow2.xplus.metadata.strategy.value.LobValueStrategy;
import org.ow2.xplus.metadata.strategy.value.TemporalValueStrategy;
import org.ow2.xplus.metadata.strategy.version.NumberVersionStrategy;
import org.ow2.xplus.metadata.strategy.version.TimestampVersionStrategy;
import org.ow2.xplus.operator.accessor.BasicAccessor;
import org.ow2.xplus.operator.accessor.DirectCollectionAccessor;
import org.ow2.xplus.operator.accessor.EmbeddedAccessor;
import org.ow2.xplus.operator.accessor.ForeignKeyAccessor;
import org.ow2.xplus.operator.accessor.RelationTableAccessor;
import org.ow2.xplus.util.ClassStrategy;
import org.ow2.xplus.util.NamingStrategy;

/**
 * member annotation processor Will read the annotation from class members,
 * apply the principles of the EJB3 spec and produces the PKUAS or-mapping
 * metadata
 * @author Lin Liang
 */

public class MemberAnnotationProcessor {

    private ClassMetaData classMetaData = null;

    private Member member = null;

    private FieldMetaData fieldMetaData = null;

    private ValueMetaData valueMetaData = null;

    private RdbTable primaryTable = null;

    private UnitMetaData unitMetaData;

    /**
     * constructor
     * @param classMetaData the classmeta of the entity class
     * @param member the class member to be processed
     */
    public MemberAnnotationProcessor(String unitName, ClassMetaData classMetaData, Member member) {
        unitMetaData = MetaDataRepository.getUnitMetaData(unitName);
        this.classMetaData = classMetaData;
        this.member = member;

        if (!isTransient()) {
            fieldMetaData = classMetaData.getFieldMetaData(member.getPropertyName());
            if (fieldMetaData.getType() == null)
                fieldMetaData.setType(member.getParameterizedType());
            if (fieldMetaData.getValue() == null)
                fieldMetaData.setValue(new ValueMetaData());
            valueMetaData = fieldMetaData.getValue();
        }
        primaryTable = unitMetaData.getTable(classMetaData.getPrimaryTableName());
    }

    /**
     * @return true if the property or field is not persistent.
     */
    private boolean isTransient() {
        return member.isAnnotationPresent(Transient.class) || member.getPropertyName().equalsIgnoreCase("serialVersionUID");
    }

    /**
     * process the or-mapping
     */
    @SuppressWarnings("unchecked")
    public void process() {
        if (isTransient())
            return;

        processSequenceGenerator();
        processTableGenerator();
        processAttributeOverrides();

        if (member.isAnnotationPresent(OneToOne.class) || member.isAnnotationPresent(OneToMany.class)
                || member.isAnnotationPresent(ManyToMany.class) || member.isAnnotationPresent(ManyToOne.class)
                || fieldMetaData.getRelation() != null) {
            return;
        }

        if (member.isAnnotationPresent(Embedded.class)) {
            processEmbedded();
            return;
        }

        if (member.isAnnotationPresent(EmbeddedId.class)) {
            processEmbeddedId();
            return;
        }

        if (member.isCollection() && (!ClassStrategy.isLobType(member.getDeclaredType()))) {
            return;
        }

        if (member.getParameterizedType().isAnnotationPresent(Embeddable.class)) {
            if (member.isAnnotationPresent(Id.class)) {
                bindEmbedded(true);
            } else {
                bindEmbedded(false);
            }
            return;
        }

        processType();
        processColumn();

        if (classMetaData.isInheritanceRoot())
            processId();

        processVersion();
    }

    /**
     * bind Collection mapping to Value metadata
     */
    private void bindCollectionValue() {
        fieldMetaData.setCollectionType(member.getDeclaredType());

        Class type = member.getDeclaredType();
        valueMetaData.setCollectionMapping(true);
        if (type.equals(Collection.class) || type.equals(Set.class))
            valueMetaData.setCollectionValue(new SetValue());
        if (type.equals(Map.class))
            valueMetaData.setCollectionValue(new MapValue());
        if (type.equals(List.class) || type.isArray())
            valueMetaData.setCollectionValue(new ListValue());

    }

    /**
     * process collection mapping
     */
    private void processCollection() {
        bindCollectionValue();
        // generate collection table
        String collectionTableName = NamingStrategy.getCollectionTableName(classMetaData.getEntityName(), member
                .getPropertyName());
        RdbTable collectionTable = unitMetaData.getTable(collectionTableName);
        // add owner Primary Key to collection table
        List<RdbColumn> joinColumns = primaryTable.getPrimaryKey().getColumns();
        for (RdbColumn joinColumn : joinColumns) {
            RdbColumn column = new RdbColumn(joinColumn);
            column.setUnique(false);
            column.setPrimaryKey(false);
            column.setTableName(collectionTableName);
            collectionTable.addColumn(column);
            collectionTable.addAttributeOverride(column.getName(), column);
        }
        // add colection value column to collection table
        RdbColumn column = new RdbColumn(member.getPropertyName());
        column.setSqlType(ClassStrategy.getSqlType(member.getDeclaredType()));
        column.setTableName(collectionTableName);
        collectionTable.addColumn(column);
        // set collection table name to value metadata
        valueMetaData.getCollectionValue().setCollectionTableName(collectionTableName);
        // add secondary table to mapping
        primaryTable.addSecondaryTableName(collectionTable.getCanonicalName());
    }

    /**
     * process
     * @TableGenerator annotation The TableGenerator annotation defines a
     *                 primary key generator that may be referenced by name when
     *                 a generator element is specified for the GeneratedValue
     *                 annotation.
     */
    private void processTableGenerator() {
        TableGenerator tableGeneratorAnn = getAnnotation(member, TableGenerator.class);
        if (tableGeneratorAnn == null)
            return;
        // prepare Generator table
        TableGeneratorStrategy tableGeneratorStrategy = new TableGeneratorStrategy(unitMetaData);

        String tableFullName = NamingStrategy.getFullTableName(tableGeneratorAnn.catalog(), tableGeneratorAnn.schema(),
                tableGeneratorAnn.name());

        RdbTable table = unitMetaData.getTable(tableFullName);

        if (!tableGeneratorAnn.catalog().equals(""))
            table.setCatalog(tableGeneratorAnn.catalog());
        if (!tableGeneratorAnn.schema().equals(""))
            table.setSchema(tableGeneratorAnn.schema());
        if (!tableGeneratorAnn.name().equals(""))
            table.setName(tableGeneratorAnn.name());

        UniqueConstraint[] uniqueConstraints = tableGeneratorAnn.uniqueConstraints();

        for (UniqueConstraint uniqueConstraint : uniqueConstraints) {
            String[] columnNames = uniqueConstraint.columnNames();
            RdbUniqueConstraint rdbUniqueConstraint = new RdbUniqueConstraint(columnNames);
            table.AddUniqueConstraint(rdbUniqueConstraint);
        }

        // set Strategy
        tableGeneratorStrategy.setTableName(tableFullName);
        tableGeneratorStrategy.setPkColumnName(tableGeneratorAnn.pkColumnName());
        tableGeneratorStrategy.setValueColumnName(tableGeneratorAnn.valueColumnName());
        tableGeneratorStrategy.setPkColumnValue(tableGeneratorAnn.pkColumnValue());
        tableGeneratorStrategy.setInitialValue(tableGeneratorAnn.initialValue());
        tableGeneratorStrategy.setAllocationSize(tableGeneratorAnn.allocationSize());
        // set IdGenerator
        RdbIdGenerator idgen = new RdbIdGenerator();
        idgen.setGeneratorStrategy(tableGeneratorStrategy);
        unitMetaData.addTableIdGen(primaryTable.getName(), idgen);

        RdbColumn pkColumn = new RdbColumn(tableGeneratorStrategy.getPkColumnName());
        pkColumn.setJavaType(String.class);
        pkColumn.setSqlType(ClassStrategy.getSqlType(String.class));
        RdbColumn valueColumn = new RdbColumn(tableGeneratorStrategy.getValueColumnName());
        valueColumn.setJavaType(Integer.class);
        valueColumn.setSqlType(ClassStrategy.getSqlType(Integer.class));

        table.addColumn(pkColumn);
        table.addColumn(valueColumn);
        pkColumn.setTableName(tableFullName);
        valueColumn.setTableName(tableFullName);
    }

    /**
     * process
     * @SequenceGenerator annotation The SequenceGenerator annotation defines a
     *                    primary key generator that may be referenced by name
     *                    when a generator element is specified for the
     *                    GeneratedValue annotation.
     */
    private void processSequenceGenerator() {
        SequenceGenerator sequenceGeneratorAnn = getAnnotation(member, SequenceGenerator.class);
        if (sequenceGeneratorAnn == null)
            return;
        // prepare Sequence Generator Strategy
        SequenceGeneratorStrategy sequenceGeneratorStrategy = new SequenceGeneratorStrategy();

        sequenceGeneratorStrategy.setName(sequenceGeneratorAnn.name());
        sequenceGeneratorStrategy.setSequenceName(sequenceGeneratorAnn.sequenceName());
        sequenceGeneratorStrategy.setInitialValue(sequenceGeneratorAnn.initialValue());
        sequenceGeneratorStrategy.setAllocationSize(sequenceGeneratorAnn.allocationSize());
        // set IdGenerator
        RdbIdGenerator idgen = new RdbIdGenerator();
        idgen.setGeneratorStrategy(sequenceGeneratorStrategy);
        unitMetaData.addTableIdGen(primaryTable.getName(), idgen);
    }

    /**
     * process
     * @Version annotation The Version annotation specifies the version field or
     *          property of an entity class that serves as its optimistic lock
     *          value.
     */
    private void processVersion() {
        if (member.isAnnotationPresent(Version.class)) {
            Class propertyType = member.getParameterizedType();
            valueMetaData.setVersionStrategy(classifyVersion(propertyType));
        }
    }

    /**
     * choose a VersionStrategy from the member type
     * @param propertyType
     * @return
     */
    private VersionStrategy classifyVersion(Class propertyType) {
        if (propertyType == int.class || propertyType == Integer.class || propertyType == short.class
                || propertyType == Short.class || propertyType == long.class || propertyType == Long.class) {
            return new NumberVersionStrategy();
        } else if (propertyType == Timestamp.class) {
            return new TimestampVersionStrategy();
        }

        MappingException me = new MappingException("type is not supported for version properties.",
                MappingException.CLASS_NOT_EMBEDDABLE);
        me.showMessage();
        throw me;
    }

    /**
     * process memver type
     */
    private void processType() {
        processBasic();
        processLob();
        processTemporal();
        processEnumerated();
    }

    /**
     * process
     * @Enumerated annotation An Enumerated annotation specifies that a
     *             persistent property or field should be persisted as a
     *             enumerated type.
     */
    @SuppressWarnings("unchecked")
    private void processEnumerated() {
        Enumerated enumeratedAnn = getAnnotation(member, Enumerated.class);

        if (!member.isEnum() && enumeratedAnn == null)
            return;
        EnumType enumType = EnumType.ORDINAL;
        if (enumeratedAnn != null) {
            enumType = enumeratedAnn.value();
        }

        // set Enumerated ValueStrategy
        EnumeratedValueStrategy enumeratedValueStrategy = new EnumeratedValueStrategy();
        Class propertyType = member.getParameterizedType();
        enumeratedValueStrategy.setValueType(propertyType);

        if (enumType.equals(EnumType.ORDINAL)) {
            enumeratedValueStrategy.setOrdinal(true);
        } else {
            enumeratedValueStrategy.setOrdinal(false);
        }
        // set to value MetaData
        valueMetaData.setValueStrategy(enumeratedValueStrategy);
    }

    /**
     * process
     * @Temporal annotation A Temporal annotation specifies that a persistent
     *           property or field should be persisted as a temporal type.
     */
    private void processTemporal() {
        Temporal temporalAnn = getAnnotation(member, Temporal.class);
        if (temporalAnn == null)
            return;
        // set Temporal ValueStrategy
        TemporalValueStrategy temporalValueStrategy = new TemporalValueStrategy();

        temporalValueStrategy.setDataType(temporalAnn.value());

        valueMetaData.setValueStrategy(temporalValueStrategy);
    }

    /**
     * process
     * @Lob annotation A Lob annotation specifies that a persistent property or
     *      field should be persisted as a large object to adatabase-supported
     *      large object type.
     */
    private void processLob() {
        Lob lobAnn = getAnnotation(member, Lob.class);
        if (lobAnn == null)
            return;

        LobValueStrategy lobValueStrategy = new LobValueStrategy();
        lobValueStrategy.setDataType(member.getParameterizedType());
        valueMetaData.setValueStrategy(lobValueStrategy);
    }

    /**
     * The Basic annotation is the simplest type of mapping to a database
     * column.
     */
    private void processBasic() {
        Basic basicAnn = getAnnotation(member, Basic.class);
        if (basicAnn != null) {
            fieldMetaData.setFetch(basicAnn.fetch());
            fieldMetaData.setOptional(basicAnn.optional());
        }
    }

    /**
     * process
     * @Id annotation The Id annotation specifies the primary key property or
     *     field of an entity.
     */
    private void processId() {
        Id idAnn = getAnnotation(member, Id.class);

        if (idAnn != null) {
            // build primary key
            RdbPrimaryKey primaryKey = primaryTable.getPrimaryKey();
            if (primaryKey == null) {
                primaryKey = new RdbPrimaryKey();
                primaryTable.setPrimaryKey(primaryKey);
            }
            // build column
            if (!member.isAnnotationPresent(Column.class)) {
                buildDefaultColumn();
            }
            RdbColumn column = valueMetaData.getSimpleValue().getColumns().get(0);
            column.setPrimaryKey(true);
            primaryTable.getColumn(column.getName()).setPrimaryKey(true);
            primaryKey.addColumn(column);
            primaryTable.setPrimaryKey(primaryKey);
            processGeneratedValue(column);
            fieldMetaData.setId(true);

            BasicAccessor accessor = new BasicAccessor(classMetaData, fieldMetaData);
            fieldMetaData.setAccessor(accessor);
        }
    }

    /**
     * process
     * @GeneratedValue annotation The GeneratedValue annotation provides for the
     *                 specification of generation strategies for the values of
     *                 primary keys.
     */
    private void processGeneratedValue(RdbColumn column) {
        GeneratedValue generatedValue = getAnnotation(member, GeneratedValue.class);
        if (generatedValue != null) {
        	fieldMetaData.setGeneratedID(true);
            RdbPrimaryKey primaryKey = primaryTable.getPrimaryKey();
            // define GenerationType
            RdbIdGenerator idgen = null;
            GenerationType strategy = generatedValue.strategy();
            if (strategy == GenerationType.TABLE) {
                String generatorName = generatedValue.generator().equals("") ? UnitMetaData.DEFAULT_TABLE_GENERATOR_NAME
                        : generatedValue.generator();
                idgen = unitMetaData.getTableIdGen(generatorName);
            } else if (strategy == GenerationType.SEQUENCE) {
                String generatorName = generatedValue.generator().equals("") ? UnitMetaData.DEFAULT_SEQUENCE_GENERATOR_NAME
                        : generatedValue.generator();
                idgen = unitMetaData.getSequenceIdGen(generatorName);
            } else if (strategy == GenerationType.IDENTITY) {
                idgen = new RdbIdGenerator();
                idgen.setGeneratorStrategy(IdentityGeneratorStrategy.getInstance());
                primaryTable.getColumn(column.getName()).setAutoIncrement(true);
                fieldMetaData.getValue().getSimpleValue().getColumns().get(0).setAutoIncrement(true);
            } else {
                strategy = GenerationType.AUTO;
                idgen = new RdbIdGenerator();
                idgen.setGeneratorStrategy(IdentityGeneratorStrategy.getInstance());
                primaryTable.getColumn(column.getName()).setAutoIncrement(true);
                fieldMetaData.getValue().getSimpleValue().getColumns().get(0).setAutoIncrement(true);
            }

            primaryKey.setIdgen(idgen);
            primaryTable.setPrimaryKey(primaryKey);
        }
        else{
        	fieldMetaData.setGeneratedID(false);
        }
    }

    /**
     * build RdbColumn on the persistence field that don't have any annotation
     */
    private void buildDefaultColumn() {
        String columnName = member.getPropertyName();
        RdbColumn column = new RdbColumn(columnName);
        column.setJavaType(member.getDeclaredType());
        String sqlType = ClassStrategy.getSqlType(member.getDeclaredType());
        if (sqlType == null)
            sqlType = ClassStrategy.getSqlType(valueMetaData.getValueStrategy().getDataType());
        column.setSqlType(sqlType);
        column.setTableName(classMetaData.getPrimaryTableName());
        primaryTable.addColumn(column);
        // bind to value metadata
        SimpleValue value = new SimpleValue();
        value.addColumn(column);
        valueMetaData.setSimpleValue(value);

        if (valueMetaData.getValueStrategy() == null) {
            BasicValueStrategy basicValueStrategy = new BasicValueStrategy();
            basicValueStrategy.setDataType(column.getJavaType());
            valueMetaData.setValueStrategy(basicValueStrategy);
        }
    }

    /**
     * process
     * @AttributeOverrides annotation The mappings of multiple properties or
     *                     fields may be overridden.
     */
    private void processAttributeOverrides() {
        AttributeOverrides attributeOverridesAnn = getAnnotation(member, AttributeOverrides.class);
        if (attributeOverridesAnn != null) {
            AttributeOverride[] attributeOverrideAnns = attributeOverridesAnn.value();
            for (AttributeOverride attributeOverrideAnn : attributeOverrideAnns) {
                processAttributeOverride(attributeOverrideAnn);
            }
        }

        AttributeOverride attributeOverrideAnn = getAnnotation(member, AttributeOverride.class);
        if (attributeOverrideAnn != null) {
            processAttributeOverride(attributeOverrideAnn);
        }
    }

    /**
     * process
     * @AttributeOverride annotation The AttributeOverride annotation is used to
     *                    override the mapping of a property or field.
     * @param attributeOverrideAnn the annotation to be processed
     */
    private void processAttributeOverride(AttributeOverride attributeOverrideAnn) {
        String fieldName = attributeOverrideAnn.name();
        Column columnAnn = attributeOverrideAnn.column();

        RdbColumn column = new RdbColumn(columnAnn.name());
        column.setUnique(columnAnn.unique());
        column.setNullable(columnAnn.nullable());
        column.setInsertable(columnAnn.insertable());
        column.setUpdatable(columnAnn.updatable());
        column.setColumnDefinition(columnAnn.columnDefinition());
        column.setLength(new Integer(columnAnn.length()));
        column.setPrecision(new Integer(columnAnn.precision()));
        column.setScale(new Integer(columnAnn.scale()));
        column.setTableName(columnAnn.table());

        valueMetaData.addAttributeOverride(fieldName, column);
    }

    /**
     * process
     * @Column annotation The Column annotation is used to specify a mapped
     *         column for a persistent property or field.
     */
    private void processColumn() {
        Column columnAnn = getAnnotation(member, Column.class);
        SimpleValue value = new SimpleValue();
        RdbColumn column = null;

        if (columnAnn != null) {
            // if there is @Column annotation
            String columnName = member.getPropertyName();
            if (!columnAnn.name().equals("")) {
                primaryTable.addAttributeOverride(columnName, new RdbColumn(columnAnn.name()));
                columnName = columnAnn.name();
            }
            // build column
            column = primaryTable.getColumn(columnName) == null ? new RdbColumn(columnName) : primaryTable.getColumn(columnName);

            column.setUnique(columnAnn.unique());
            column.setNullable(columnAnn.nullable());
            column.setInsertable(columnAnn.insertable());
            column.setUpdatable(columnAnn.updatable());
            column.setColumnDefinition(columnAnn.columnDefinition());
            column.setLength(new Integer(columnAnn.length()));
            column.setPrecision(new Integer(columnAnn.precision()));
            column.setScale(new Integer(columnAnn.scale()));
            column.setCanonicalFieldName(fieldMetaData.getFieldName());
            column.setJavaType(member.getDeclaredType());

            String sqlType = ClassStrategy.getSqlType(member.getDeclaredType());
            if (sqlType == null)
                sqlType = ClassStrategy.getSqlType(valueMetaData.getValueStrategy().getDataType());
            column.setSqlType(sqlType);
            // add to table
            String tableName = columnAnn.table().equals("") ? classMetaData.getPrimaryTableName() : columnAnn.table();
            column.setTableName(tableName);
            unitMetaData.getTable(tableName).addColumn(column);
        } else {
            // no @Column annotation, build defaule column
            String columnName = member.getPropertyName();
            column = primaryTable.getColumn(columnName) == null ? new RdbColumn(columnName) : primaryTable.getColumn(columnName);
            column.setTableName(primaryTable.getCanonicalName());
            column.setCanonicalFieldName(fieldMetaData.getFieldName());
            primaryTable.addColumn(column);

        }

        if (valueMetaData.getValueStrategy() == null) {
            BasicValueStrategy basicValueStrategy = new BasicValueStrategy();
            basicValueStrategy.setDataType(member.getParameterizedType());
            valueMetaData.setValueStrategy(basicValueStrategy);
        }

        Class javaType = member.getParameterizedType();
        column.setJavaType(javaType);
        String sqlType = ClassStrategy.getSqlType(valueMetaData.getValueStrategy().getDataType());
        column.setSqlType(sqlType);
        value.addColumn(column);
        valueMetaData.setSimpleValue(value);

        BasicAccessor accessor = new BasicAccessor(classMetaData, fieldMetaData);
        fieldMetaData.setAccessor(accessor);
    }

    /**
     * process entity classes relationship
     */
    public void processRelationship() {
        if (isTransient())
            return;

        Relationship relation = null;

        if (member.isAnnotationPresent(OneToOne.class)) {
            OneToOne oneToOne = member.getAnnotation(OneToOne.class);
            relation = new OneToOneRelation(member, oneToOne);
            setRelationship(relation);
            processOneToOneRelation(relation);
        } else if (member.isAnnotationPresent(ManyToOne.class)) {
            ManyToOne manyToOne = member.getAnnotation(ManyToOne.class);
            relation = new ManyToOneRelation(member, manyToOne);
            setRelationship(relation);
            processManyToOneRelation(relation);
        } else if (member.isAnnotationPresent(ManyToMany.class)) {
            ManyToMany manyToMany = member.getAnnotation(ManyToMany.class);
            relation = new ManyToManyRelation(member, manyToMany);
            setRelationship(relation);
            processManyToManyRelation(relation);
        } else if (member.isAnnotationPresent(OneToMany.class)) {
            OneToMany oneToMany = member.getAnnotation(OneToMany.class);
            relation = new OneToManyRelation(member, oneToMany);
            setRelationship(relation);
            processOneToManyRelation(relation);
        } else {
            // if no relation annotation
            // member is collection
            if (member.isCollection()) {
                // return if lob type
                if (ClassStrategy.isLobType(member.getDeclaredType()))
                    return;
                // is a SerializedType
                if (ClassStrategy.isSerializedType(member.getParameterizedType()) || member.getParameterizedType().isEnum()) {
                    processCollection();
                    DirectCollectionAccessor accessor = new DirectCollectionAccessor(classMetaData, fieldMetaData);
                    fieldMetaData.setAccessor(accessor);
                } else {
                    // is a collecton of entities
                    relation = new OneToManyRelation(member);
                    setRelationship(relation);
                    processOneToManyRelation(relation);
                }
            }
        }
    }

    /**
     * set relationship to field MetaData
     * @param relation
     */
    private void setRelationship(Relationship relation) {
        fieldMetaData.setRelation(relation);
        fieldMetaData.setFetch(relation.getFetch());
        fieldMetaData.setOptional(relation.isOptional());
    }

    /**
     * build the relationship at the target side
     */
    public void processInverseRelationship() {
        if (isTransient())
            return;

        Relationship relationship = fieldMetaData.getRelation();
        if (relationship == null)
            return;
        RdbRelation relation = relationship.buildInverseRelation(classMetaData.getUnitName(), relationship.getTargetEntity()
                .getName());
        if (relation != null) {
            primaryTable.addRelation(member.getPropertyName(), relation);
            if (relation.getOwnerTableName() == null)
                relation.setOwnerTableName(primaryTable.getName());
        }
    }

    /**
     * check if the table has JoinColumn
     * @return true if the table has JoinColumn
     */
    private boolean hasJoinColumn() {
        boolean result = member.isAnnotationPresent(JoinColumn.class) || member.isAnnotationPresent(JoinColumns.class)
                || member.isAnnotationPresent(PrimaryKeyJoinColumn.class)
                || member.isAnnotationPresent(PrimaryKeyJoinColumns.class);
        return result;
    }

    /**
     * process
     * @ManyToOne Relation
     * @param relation
     */
    private void processManyToOneRelation(Relationship relation) {
        processJoinColumns();
        if (!hasJoinColumn()) {
            buildDefaultJoinColumn();
        }
        ForeignKeyAccessor accessor = new ForeignKeyAccessor(classMetaData, fieldMetaData);
        fieldMetaData.setAccessor(accessor);
    }

    /**
     * process
     * @ManyToMany Relation
     * @param relation
     */
    private void processManyToManyRelation(Relationship relation) {
        bindCollectionValue();
        processOrderBy();
        processMapKey();

        if (relation.getMappedBy().equals("")) {
            processJoinTable();
        }

        RelationTableAccessor accessor = new RelationTableAccessor(classMetaData, fieldMetaData);
        fieldMetaData.setAccessor(accessor);
    }

    /**
     * process
     * @OneToMany Relation
     * @param relation
     */
    private void processOneToManyRelation(Relationship relation) {
        bindCollectionValue();
        processOrderBy();
        processMapKey();
        if (relation.getMappedBy().equals("")) {
            processJoinTable();
        }
        RelationTableAccessor accessor = new RelationTableAccessor(classMetaData, fieldMetaData);
        fieldMetaData.setAccessor(accessor);
    }

    /**
     * process
     * @OneToOne Relation
     * @param relation
     */
    private void processOneToOneRelation(Relationship relation) {
        // if it's not a inverse relation
        if (relation.getMappedBy().equals("")) {
            processJoinColumns();
            if (primaryTable.getRelation(member.getPropertyName()) == null) {
                processPrimaryKeyJoinColumns();
            }

            if (!hasJoinColumn()) {
                buildDefaultJoinColumn();
            }
            ForeignKeyAccessor accessor = new ForeignKeyAccessor(classMetaData, fieldMetaData);
            fieldMetaData.setAccessor(accessor);
        }
    }

    /**
     * add Primary key as Default JoinColumn if is not specified
     */
    private void buildDefaultJoinColumn() {
        RdbForeignKey foreignKey = new RdbForeignKey();

        String referencedEntityName = fieldMetaData.getRelation().getTargetEntity().getName();
        foreignKey.setReferencedEntityName(referencedEntityName);
        String referencedTableName = unitMetaData.getClassMetaData(referencedEntityName).getPrimaryTableName();
        foreignKey.setReferencedTableName(referencedTableName);

        RdbTable referencedTable = unitMetaData.getTable(referencedTableName);
        RdbPrimaryKey referencedPrimaryKey = referencedTable.getPrimaryKey();
        List<RdbColumn> referencedPrimaryColumns = referencedPrimaryKey.getColumns();

        if (referencedPrimaryColumns.isEmpty()) {
            foreignKey.setJoinPrimaryKey(referencedPrimaryKey);
        }

        for (RdbColumn referencedPrimaryColumn : referencedPrimaryColumns) {

            String joinColumnName = NamingStrategy.getJoinColumnName(fieldMetaData.getFieldName(), referencedPrimaryColumn
                    .getName());
            RdbColumn joinColumn = new RdbColumn(referencedPrimaryColumn);
            joinColumn.setName(joinColumnName);
            joinColumn.setTableName(primaryTable.getCanonicalName());
            joinColumn.setCanonicalFieldName(NamingStrategy.getCanonicalFieldName(member.getPropertyName(),
                    referencedPrimaryColumn.getCanonicalFieldName()));
            joinColumn.setUnique(false);
            joinColumn.setPrimaryKey(false);

            foreignKey.addcolumnMaping(joinColumnName, referencedPrimaryColumn.getName());
            primaryTable.addColumn(joinColumn);
        }

        foreignKey.setOwnerTableName(primaryTable.getCanonicalName());
        primaryTable.addRelation(member.getPropertyName(), foreignKey);
    }

    /**
     * process
     * @PrimaryKeyJoinColumns annotation Composite foreign keys are supported by
     *                        means of the PrimaryKeyJoinColumns annotation.
     */
    private void processPrimaryKeyJoinColumns() {
        RdbForeignKey foreignKey = new RdbForeignKey();

        PrimaryKeyJoinColumns primaryKeyJoinColumnsAnn = getAnnotation(member, PrimaryKeyJoinColumns.class);
        if (primaryKeyJoinColumnsAnn != null) {
            PrimaryKeyJoinColumn[] primaryKeyJoinColumnAnns = primaryKeyJoinColumnsAnn.value();
            for (PrimaryKeyJoinColumn primaryKeyJoinColumnAnn : primaryKeyJoinColumnAnns) {
                processPrimaryKeyJoinColumn(primaryKeyJoinColumnAnn, foreignKey);
            }
        }

        PrimaryKeyJoinColumn primaryKeyJoinColumnAnn = getAnnotation(member, PrimaryKeyJoinColumn.class);
        if (primaryKeyJoinColumnAnn != null) {
            processPrimaryKeyJoinColumn(primaryKeyJoinColumnAnn, foreignKey);
        }

        primaryTable.addRelation(member.getPropertyName(), foreignKey);
        foreignKey.setOwnerTableName(primaryTable.getCanonicalName());
    }

    /**
     * process
     * @PrimaryKeyJoinColumn annotation The PrimaryKeyJoinColumn annotation
     *                       specifies a primary key column that is used as a
     *                       foreign key to join to another table.
     * @param primaryKeyJoinColumnAnn
     * @PrimaryKeyJoinColumn annotation
     * @param foreignKey the foreign key to bind to
     */
    private void processPrimaryKeyJoinColumn(PrimaryKeyJoinColumn primaryKeyJoinColumnAnn, RdbForeignKey foreignKey) {
        String referencedEntityName = fieldMetaData.getRelation().getTargetEntity().getName();
        foreignKey.setReferencedEntityName(referencedEntityName);
        String referencedTableName = unitMetaData.getClassMetaData(referencedEntityName).getPrimaryTableName();
        foreignKey.setReferencedTableName(referencedTableName);
        // build join column
        RdbTable referencedTable = unitMetaData.getTable(referencedTableName);
        RdbPrimaryKey referencedPrimaryKey = referencedTable.getPrimaryKey();
        String referencedPrimaryColumnName = primaryKeyJoinColumnAnn.referencedColumnName().equals("") ? referencedPrimaryKey
                .getColumns().get(0).getName() : primaryKeyJoinColumnAnn.referencedColumnName();

        RdbColumn referencedPrimaryColumn = referencedTable.getColumn(referencedPrimaryColumnName);
        String joinColumnName = primaryKeyJoinColumnAnn.name().equals("") ? NamingStrategy.getJoinColumnName(fieldMetaData
                .getFieldName(), referencedPrimaryColumnName) : primaryKeyJoinColumnAnn.name();
        RdbColumn joinColumn = new RdbColumn(referencedPrimaryColumn);
        joinColumn.setName(joinColumnName);
        joinColumn.setColumnDefinition(primaryKeyJoinColumnAnn.columnDefinition());
        joinColumn.setTableName(primaryTable.getCanonicalName());
        joinColumn.setCanonicalFieldName(NamingStrategy.getCanonicalFieldName(member.getPropertyName(), referencedPrimaryColumn
                .getCanonicalFieldName()));
        // add to mapping
        foreignKey.addcolumnMaping(joinColumnName, referencedPrimaryColumn.getName());
        primaryTable.addColumn(joinColumn);
    }

    /**
     * process
     * @MapKey annotation The MapKey annotation is used to specify the map key
     *         for associations of type java.util.Map.
     */
    private void processMapKey() {
        MapKey mapKeyAnn = getAnnotation(member, MapKey.class);

        if (mapKeyAnn != null) {
            CollectionValue collection = valueMetaData.getCollectionValue();

            String mappedPropertyName = mapKeyAnn.name().equals("") ? primaryTable.getPrimaryKey().getColumns().get(0).getName()
                    : mapKeyAnn.name();

            collection.setMappedPropertyName(mappedPropertyName);
        }

    }

    /**
     * process
     * @OrderBy annotation The OrderBy annotation specifies the ordering of the
     *          elements of a collection valued association at the point when
     *          the association is retrieved.
     */
    private void processOrderBy() {
        OrderBy orderByAnn = getAnnotation(member, OrderBy.class);

        if (orderByAnn != null) {
            String orderbyList = orderByAnn.value();
            String propertyName;
            String ordering = CollectionValue.ASC;

            CollectionValue collection = valueMetaData.getCollectionValue();

            if (orderbyList.equals("")) {
                // If the ordering element is not specified, ordering by the
                // primary key of the associated entity is assumed.
                propertyName = primaryTable.getPrimaryKey().getColumns().get(0).getName();

                collection.addOrderBy(propertyName, ordering);
            } else {
                // orderby_list::= orderby_item [,orderby_item]*
                StringTokenizer orderItem = new StringTokenizer(orderbyList, ",");

                if (orderItem.hasMoreTokens()) {
                    // orderby_item::= property_or_field_name [ASC | DESC]
                    StringTokenizer token = new StringTokenizer(orderItem.nextToken(), " ");

                    propertyName = token.nextToken();

                    if (token.hasMoreTokens()) {
                        ordering = token.nextToken();
                    }

                    collection.addOrderBy(propertyName, ordering);
                }
            }
        }
    }

    /**
     * process
     * @JoinColumns annotation Composite foreign keys are supported by means of
     *              the JoinColumns annotation.
     */
    private void processJoinColumns() {
        // build foreign key
        RdbForeignKey foreignKey = new RdbForeignKey();

        String referencedEntityName = fieldMetaData.getRelation().getTargetEntity().getName();
        foreignKey.setReferencedEntityName(referencedEntityName);
        String referencedTableName = unitMetaData.getClassMetaData(referencedEntityName).getPrimaryTableName();
        foreignKey.setReferencedTableName(referencedTableName);
        // process JoinColumns
        JoinColumns joinColumnsAnn = getAnnotation(member, JoinColumns.class);
        if (joinColumnsAnn != null) {
            int index = 0;
            for (JoinColumn joinColumnAnn : joinColumnsAnn.value()) {
                buildForeignKeyJoinColumn(foreignKey, joinColumnAnn, index++);
            }
        }

        JoinColumn joinColumnAnn = getAnnotation(member, JoinColumn.class);
        if (joinColumnAnn != null) {
            buildForeignKeyJoinColumn(foreignKey, joinColumnAnn, 0);
        }
        // add to mapping
        primaryTable.addRelation(member.getPropertyName(), foreignKey);
    }

    /**
     * build JoinColumn used in ForeignKey
     * @param foreignKey
     * @param joinColumnAnn
     * @JoinColumn annotation to be processed
     * @param index the index of the
     * @JoinColumn in
     * @JoinColumns, default 0
     */
    private void buildForeignKeyJoinColumn(RdbForeignKey foreignKey, JoinColumn joinColumnAnn, int index) {
        RdbColumn referencedColumn = null;
        RdbTable referencedTable = unitMetaData.getTable(foreignKey.getReferencedTableName());
        if (joinColumnAnn.referencedColumnName().equals("")) {
            // embedded id causes null
            if (!referencedTable.getPrimaryKey().getColumns().isEmpty()) {
                String columnName = referencedTable.getPrimaryKey().getColumns().get(index).getName();
                referencedColumn = referencedTable.getColumn(columnName);
            } else {
                // add Join PrimaryKey to foreignKey
                foreignKey.setJoinPrimaryKey(referencedTable.getPrimaryKey());
                foreignKey.addJoinPrimaryKeyColumn(index, joinColumnAnn);
                return;
            }
        } else {
            String referencedColumnName = joinColumnAnn.referencedColumnName();
            referencedColumn = referencedTable.getColumn(referencedColumnName);
            if (referencedColumn == null)
                referencedColumn = new RdbColumn(referencedColumnName);
        }
        // build JoinColumn mapping
        String fieldName = fieldMetaData.getFieldName();
        String ownerTableName = joinColumnAnn.table();
        if (ownerTableName.equals(""))
            ownerTableName = classMetaData.getPrimaryTableName();
        foreignKey.bindJoinColumn(unitMetaData.getUnitName(), fieldName, ownerTableName, referencedColumn, joinColumnAnn);
        foreignKey.setOwnerTableName(ownerTableName);
    }

    /**
     * process
     * @JoinTable annotation The JoinTable annotation is used in the mapping of
     *            associations.
     */
    @SuppressWarnings("unchecked")
    private void processJoinTable() {
        JoinTable joinTableAnn = getAnnotation(member, JoinTable.class);

        String referencedEntityName = fieldMetaData.getRelation().getTargetEntity().getName();
        ClassMetaData referencedEntityMetaData = unitMetaData.getClassMetaData(referencedEntityName);
        String referencedEntityTableName = referencedEntityMetaData.getPrimaryTableName();
        // get the name of joinTable
        String joinTableName;

        if (joinTableAnn == null || joinTableAnn.name().equals("")) {
            joinTableName = NamingStrategy.getJoinTableName(primaryTable.getCanonicalName(), referencedEntityTableName);
        } else {
            joinTableName = joinTableAnn.name();
        }
        // build joinTable
        RdbTable table = unitMetaData.getTable(joinTableName);
        JoinColumn[] joinColumnAnns = {};
        JoinColumn[] inverseJoinColumns = {};

        if (joinTableAnn != null) {
            // if exists @JoinTable
            if (!joinTableAnn.catalog().equals(""))
                table.setCatalog(joinTableAnn.catalog());
            if (!joinTableAnn.schema().equals(""))
                table.setSchema(joinTableAnn.schema());
            if (!joinTableAnn.name().equals(""))
                table.setName(joinTableAnn.name());

            UniqueConstraint[] uniqueConstraints = joinTableAnn.uniqueConstraints();

            for (UniqueConstraint uniqueConstraint : uniqueConstraints) {
                String[] columnNames = uniqueConstraint.columnNames();
                RdbUniqueConstraint rdbUniqueConstraint = new RdbUniqueConstraint(columnNames);
                table.AddUniqueConstraint(rdbUniqueConstraint);
            }

            joinColumnAnns = joinTableAnn.joinColumns();
            inverseJoinColumns = joinTableAnn.inverseJoinColumns();
        }

        RdbRelationTable relationTable = new RdbRelationTable();
        relationTable.setRelationTable(table);
        relationTable.setReferencedEntityName(referencedEntityName);
        relationTable.setReferencedTableName(referencedEntityTableName);

        // add join Columns
        if (joinColumnAnns.length != 0) {
            int index = 0;
            for (JoinColumn joinColumnAnn : joinColumnAnns) {
                buildJoinTableJoinColumn(relationTable, joinColumnAnn, index++, false);
            }
        } else {
            // add PrimaryKey as joinColumn
            RdbPrimaryKey primaryKey = primaryTable.getPrimaryKey();
            buildJoinTableJoinColumn(relationTable, primaryKey, false);
        }
        // add inverse join Columns
        if (inverseJoinColumns.length != 0) {
            for (JoinColumn joinColumnAnn : inverseJoinColumns) {
                buildJoinTableJoinColumn(relationTable, joinColumnAnn, 0, true);
            }
        } else {
            // add referenced entity PrimaryKey as joinColumn
            if (member.getParameterizedType().isAnnotationPresent(Entity.class)) {
                RdbPrimaryKey referencedPrimaryKey = unitMetaData.getTable(referencedEntityTableName).getPrimaryKey();
                buildJoinTableJoinColumn(relationTable, referencedPrimaryKey, true);
            }
        }

        primaryTable.addRelation(member.getPropertyName(), relationTable);
        relationTable.setOwnerTableName(primaryTable.getName());
    }

    /**
     * build JoinColumns in join relation table
     * @param relationTable the relation table
     * @param joinColumn the join column
     * @param inverse if it is a inverse JoinColumn
     */
    private void buildJoinTableJoinColumn(RdbRelationTable relationTable, RdbPrimaryKey joinPrimaryKey, boolean inverse) {
        List<RdbColumn> joinColumns = joinPrimaryKey.getColumns();
        if (joinColumns.size() > 0) {
            for (RdbColumn joinColumn : joinColumns) {
                String relationColumnName = null;
                // get column name
                if (inverse) {
                    relationColumnName = NamingStrategy.getJoinColumnName(fieldMetaData.getFieldName(), joinColumn.getName());
                } else {
                    relationColumnName = NamingStrategy.getJoinColumnName(classMetaData.getEntityName(), joinColumn.getName());
                }
                // build join column
                RdbColumn column = new RdbColumn(joinColumn);
                column.setName(relationColumnName);
                column.setUnique(false);
                column.setPrimaryKey(false);

                RdbTable table = relationTable.getRelationTable();
                column.setTableName(table.getCanonicalName());

                table.addColumn(column);
                // add to mapping
                if (inverse) {
                    relationTable.addTargetRelationMapping(relationColumnName, joinColumn.getName());
                } else {
                    relationTable.addSourceRelationMapping(relationColumnName, joinColumn.getName());
                }
            }
        } else {
            if (inverse) {
                relationTable.setTargetJoinPrimaryKey(joinPrimaryKey);
            } else {
                relationTable.setSourceJoinPrimaryKey(joinPrimaryKey);
            }
        }
    }

    /**
     * build JoinColumns in join relation table from annotation
     * @param relationTable the relation table
     * @param joinColumnAnn the
     * @JoinColumn annotation to be processed
     * @param index the index of the
     * @JoinColumn in
     * @JoinColumns, default 0
     * @param inverse true if it is an inverse column
     */
    private void buildJoinTableJoinColumn(RdbRelationTable relationTable, JoinColumn joinColumnAnn, int index, boolean inverse) {
        RdbColumn referencedColumn = null;
        // get referenced table
        String referencedTableName = null;
        if (inverse) {
            referencedTableName = relationTable.getReferencedTableName();
        } else {
            referencedTableName = classMetaData.getPrimaryTableName();
        }
        RdbTable referencedTable = unitMetaData.getTable(referencedTableName);

        if (joinColumnAnn.referencedColumnName().equals("")) {
            // embedded id may cause null
            if (!referencedTable.getPrimaryKey().getColumns().isEmpty()) {
                referencedColumn = referencedTable.getPrimaryKey().getColumns().get(index);
            } else {
                if (inverse) {
                    relationTable.setTargetJoinPrimaryKey(referencedTable.getPrimaryKey());
                    relationTable.addTargetJoinPrimaryKeyColumn(index, joinColumnAnn);
                } else {
                    relationTable.setSourceJoinPrimaryKey(referencedTable.getPrimaryKey());
                    relationTable.addSourceJoinPrimaryKeyColumn(index, joinColumnAnn);
                }
                return;
            }
        } else {
            String referencedColumnName = joinColumnAnn.referencedColumnName();
            referencedColumn = referencedTable.getColumn(referencedColumnName);
            if (referencedColumn == null)
                referencedColumn = new RdbColumn(referencedColumnName);
        }

        String fieldName = fieldMetaData.getFieldName();
        relationTable.bindJoinColumn(fieldName, referencedColumn, joinColumnAnn, inverse);
    }

    /**
     * process
     * @Embedded annotation The Embedded annotation is used to specify a
     *           persistent field or property of an entity whose value is an
     *           instance of an embeddable class.
     */
    private void processEmbedded() {
        Embedded embeddedAnn = getAnnotation(member, Embedded.class);
        if (embeddedAnn == null)
            return;
        bindEmbedded(false);
    }

    /**
     * process
     * @EmbeddedId annotation The EmbeddedId annotation is applied to a
     *             persistent field or property of an entity class or mapped
     *             superclass to denote a composite primary key that is an
     *             embeddable class.
     */
    private void processEmbeddedId() {
        EmbeddedId embeddedIdAnn = getAnnotation(member, EmbeddedId.class);
        if (embeddedIdAnn == null)
            return;
        bindEmbedded(true);
    }

    /**
     * bind embedded class to entity table
     * @param isId true if it's EmbeddedId
     */
    @SuppressWarnings("unchecked")
    private void bindEmbedded(boolean isId) {
        Class embeddedClass = member.getParameterizedType();
        // whether it is Embeddable
        if (!embeddedClass.isAnnotationPresent(Embeddable.class)) {
            if (embeddedClass.isAnnotationPresent(Entity.class) || embeddedClass.isAnnotationPresent(MappedSuperclass.class)) {
                MappingException me = new MappingException("Embedded entity is not an instance of an embeddable class.",
                        MappingException.CLASS_NOT_EMBEDDABLE);
                me.showMessage();
                throw me;
            }
        }

        SimpleValue value = new SimpleValue();
        value.setId(isId);
        value.setEmbedded(true);
        value.setEmbeddedClassName(embeddedClass.getName());
        // set to mapping
        valueMetaData.setSimpleValue(value);
        fieldMetaData.setValue(valueMetaData);

        EmbeddedAccessor accessor = new EmbeddedAccessor(classMetaData, fieldMetaData);
        fieldMetaData.setAccessor(accessor);

        if (isId) {
            fieldMetaData.setId(true);
            RdbPrimaryKey primaryKey = new RdbPrimaryKey();
            primaryKey.setPkClassName(embeddedClass.getName());
            primaryTable.setPrimaryKey(primaryKey);
        }
    }

    /**
     * sort the classes to place embeddedClass first
     * 
     * @param orderedClasses the ordered Classes
     * @param index index of the place to insert in
     */
    public void orderEmbaddable(List<Class> orderedClasses, int index) {
        if (isTransient())
            return;

        SimpleValue value = valueMetaData.getSimpleValue();
        if (value == null || !value.isEmbedded())
            return;

        Class embeddedClass = ClassStrategy.forName(value.getEmbeddedClassName());
        if (orderedClasses.contains(embeddedClass))
            return;
        orderedClasses.add(index, embeddedClass);
    }

    /**
     * bind Embedded value to mapping
     *
     */
    public void buildEmbeddedMapping() {
        if (isTransient())
            return;
        buildEmbeddedValue();
    }

    /**
     * bind Embedded value in relation to mapping
     */
    public void buildRelationEmbeddedMapping() {
        if (isTransient())
            return;
        buildRelationColumn();
    }

    /**
     * bind Embedded column to relation
     *
     */
    private void buildRelationColumn() {
        String tableName = classMetaData.getPrimaryTableName();
        primaryTable = unitMetaData.getTable(tableName);
        String fieldName = fieldMetaData.getFieldName();
        RdbRelation relation = primaryTable.getRelation(fieldName);
        if (relation != null)
            relation.bindEmbeddedColumn(unitMetaData.getUnitName(), fieldName, tableName);
    }

    /**
     * bind Embedded columns to mapping
     */
    private void buildEmbeddedValue() {
        if (valueMetaData.isCollectionMapping())
            return;
        SimpleValue value = valueMetaData.getSimpleValue();
        if (value == null)
            return;
        if (!value.isEmbedded())
            return;

        String embeddedClassName = value.getEmbeddedClassName();
        ClassMetaData embeddedClassMetaData = unitMetaData.getClassMetaData(embeddedClassName);
        String embeddedClassTableName = embeddedClassMetaData.getPrimaryTableName();
        RdbTable embeddedClassTable = unitMetaData.getTable(embeddedClassTableName);
        List<RdbColumn> embeddedClassTableColumns = embeddedClassTable.getColumns();

        for (RdbColumn embeddedClassTableColumn : embeddedClassTableColumns) {
            RdbColumn column = new RdbColumn(embeddedClassTableColumn);
            RdbColumn overrideColumnInfo = valueMetaData.getAttributeOverride(embeddedClassTableColumn.getCanonicalFieldName());
            String columnName = overrideColumnInfo == null ? "" : overrideColumnInfo.getName();
            if (columnName.equals("")) {
                //columnName = NamingStrategy.getJoinColumnName(member.getPropertyName(), embeddedClassTableColumn.getName());
                columnName = embeddedClassTableColumn.getName();
                valueMetaData.addAttributeOverride(embeddedClassTableColumn.getCanonicalFieldName(), new RdbColumn(columnName));
            }
            column.setName(NamingStrategy.getEmbeddedColumnName(member.getPropertyName(), columnName));
            column.setCanonicalFieldName(NamingStrategy.getCanonicalFieldName(member.getPropertyName(), embeddedClassTableColumn
                    .getCanonicalFieldName()));

            if (overrideColumnInfo == null || overrideColumnInfo.getTableName().equals("")) {
                primaryTable.addColumn(column);
                column.setTableName(primaryTable.getCanonicalName());
            } else {
                unitMetaData.getTable(overrideColumnInfo.getTableName()).addColumn(column);
                column.setTableName(overrideColumnInfo.getTableName());
            }

            value.addColumn(column);
            if (value.isId()) {
                column.setPrimaryKey(true);
                primaryTable.getPrimaryKey().addColumn(column);
            }
        }
    }

    /**
     * override the mapping of a property or field.
     */
    public void bindAttributeOverride() {
        if (isTransient())
            return;
        if (member.isCollection())
            return;
        if (valueMetaData.isCollectionMapping())
            return;
        SimpleValue value = valueMetaData.getSimpleValue();
        if (value == null)
            return;

        for (RdbColumn column : value.getColumns()) {
            RdbColumn overrideColumnInfo = valueMetaData.getAttributeOverride(column.getCanonicalFieldName());
            if (!(overrideColumnInfo == null)) {
                overrideColumnInfo(column, overrideColumnInfo);
            } else {
                overrideColumnInfo = primaryTable.getAttributeOverride(column.getCanonicalFieldName());
                if (!(overrideColumnInfo == null)) {
                    overrideColumnInfo(column, overrideColumnInfo);
                }
            }
        }
    }

    /**
     * override Column Information
     * 
     * @param column column to be overrided
     * @param overrideColumn override information
     */
    private void overrideColumnInfo(RdbColumn column, RdbColumn overrideColumn) {
        if (!overrideColumn.getName().equals(""))
            column.setName(overrideColumn.getName());
        if (!overrideColumn.getTableName().equals(""))
            column.setTableName(overrideColumn.getName());
    }

    private <T extends Annotation> T getAnnotation(AnnotatedElement annElement, Class<T> annotationClass) {
        T ann = null;
        if (annElement.isAnnotationPresent(annotationClass)) {
            ann = annElement.getAnnotation(annotationClass);
        }
        return ann;
    }
}
