/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.metadata;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.persistence.spi.ClassTransformer;
import javax.persistence.spi.PersistenceUnitInfo;
import javax.persistence.spi.PersistenceUnitTransactionType;
import javax.sql.DataSource;

/**
 * @author Lin Liang
 */

public class PersistenceInfo implements PersistenceUnitInfo {

    /**
     * The name of the persistence unit. Corresponds to the <name> element in
     * the persistence.xml file.
     */
    private String persistenceUnitName;

    /**
     * The fully qualified name of the persistence provider implementation
     * class. Corresponds to the <provider> element in the persistence.xml file.
     */
    private String persistenceProviderClassName;

    /**
     * The transaction type of the entity managers created by the
     * EntityManagerFactory. The transaction type corresponds to the
     * transaction-type attribute in the persistence.xml file.
     */
    private PersistenceUnitTransactionType transactionType;

    /**
     * The JTA-enabled data source to be used by the persistence provider. The
     * data source corresponds to the <jta-data-source> element in the
     * persistence.xml file or is provided at deployment or by the container.
     */
    private DataSource jtaDataSource;

    /**
     * The non-JTA-enabled data source to be used by the persistence provider
     * for accessing data outside a JTA transaction. The data source corresponds
     * to the named <non-jta-data-source> element in the persistence.xml file or
     * provided at deployment or by the container.
     */
    private DataSource nonJtaDataSource;

    /**
     * The list of mapping file names that the persistence provider must load to
     * determine the mappings for the entity classes. The mapping files must be
     * in the standard XML mapping format, be uniquely named and be
     * resource-loadable from the application classpath. Each mapping file name
     * corresponds to a <mapping-file> element in the persistence.xml file.
     */
    private List<String> mappingFileNames = new ArrayList<String>();

    /**
     * The list of JAR file URLs that the persistence provider must examine for
     * managed classes of the persistence unit. Each jar file URL corresponds to
     * a named <jar-file> element in the persistence.xml file.
     */
    private List<URL> jarFileUrls = new ArrayList<URL>();

    /**
     * The URL for the jar file or directory that is the root of the persistence
     * unit. (If the persistence unit is rooted in the WEB-INF/classes
     * directory, this will be the URL of that directory.)
     */
    private URL persistenceUnitRootUrl;

    /**
     * The list of the names of the classes that the persistence provider must
     * add it to its set of managed classes. Each name corresponds to a named
     * <class> element in the persistence.xml file.
     */
    private List<String> managedClassNames = new ArrayList<String>();

    /**
     * Whether classes in the root of the persistence unit that have not been
     * explicitly listed are to be included in the set of managed classes. This
     * value corresponds to the <exclude-unlisted-classes> element in the
     * persistence.xml file.
     */
    private boolean excludeUnlistedClasses;

    /**
     * Properties object. Each property corresponds to a <property> element in
     * the persistence.xml file
     */
    private Properties properties = new Properties();

    /**
     * ClassLoader that the provider may use to load any classes, resources, or
     * open URLs.
     */
    private ClassLoader classLoader;

    /**
     * transformer A provider-supplied transformer that the Container invokes at
     * class-(re)definition time
     */
    private List<ClassTransformer> transformers = new ArrayList<ClassTransformer>();

    /**
     * Temporary ClassLoader with same visibility as current loader
     */
    private ClassLoader newTempClassLoader;

    public PersistenceInfo() {

    }

    public PersistenceInfo(PersistenceUnitInfo info) {
        this.excludeUnlistedClasses = info.excludeUnlistedClasses();
        this.classLoader = info.getClassLoader();
        this.jarFileUrls = info.getJarFileUrls();
        this.jtaDataSource = info.getJtaDataSource();
        this.managedClassNames = info.getManagedClassNames();
        this.mappingFileNames = info.getMappingFileNames();
        this.newTempClassLoader = info.getNewTempClassLoader();
        this.nonJtaDataSource = info.getNonJtaDataSource();
        this.persistenceProviderClassName = info.getPersistenceProviderClassName();
        this.persistenceUnitName = info.getPersistenceUnitName();
        this.persistenceUnitRootUrl = info.getPersistenceUnitRootUrl();
        this.properties = info.getProperties();
        this.transactionType = info.getTransactionType();
    }

    public String getPersistenceUnitName() {
        return persistenceUnitName;
    }

    public void setPersistenceUnitName(String persistenceUnitName) {
        this.persistenceUnitName = persistenceUnitName;
    }

    public String getPersistenceProviderClassName() {
        return persistenceProviderClassName;
    }

    public void setPersistenceProviderClassName(String persistenceProviderClassName) {
        this.persistenceProviderClassName = persistenceProviderClassName;
    }

    public PersistenceUnitTransactionType getTransactionType() {
        return transactionType;
    }

    public void setTransactionType(PersistenceUnitTransactionType transactionType) {
        this.transactionType = transactionType;
    }

    public void setTransactionTypeName(String TransactionTypeName) {
        if (TransactionTypeName.equalsIgnoreCase("JTA")) {
            this.transactionType = PersistenceUnitTransactionType.JTA;
        } else {
            this.transactionType = PersistenceUnitTransactionType.RESOURCE_LOCAL;
        }
    }

    public DataSource getJtaDataSource() {
        return jtaDataSource;
    }

    public void setJtaDataSource(DataSource jtaDataSource) {
        this.jtaDataSource = jtaDataSource;
    }

    public void setJtaDataSourceName(String jtaDataSourceName) {
        DataSource jtaDataSource = null;

        try {
            InitialContext initialContext = new InitialContext();
            jtaDataSource = (DataSource) initialContext.lookup(jtaDataSourceName);
        } catch (NamingException ne) {
            // ne.printStackTrace();
        }

        this.jtaDataSource = jtaDataSource;
    }

    public DataSource getNonJtaDataSource() {
        return nonJtaDataSource;
    }

    public void setNonJtaDataSource(DataSource nonJtaDataSource) {
        this.nonJtaDataSource = nonJtaDataSource;
    }

    public void setNonJtaDataSourceName(String nonJtaDataSourceName) {
        DataSource nonJtaDataSource = null;

        try {
            InitialContext initialContext = new InitialContext();
            jtaDataSource = (DataSource) initialContext.lookup(nonJtaDataSourceName);
        } catch (NamingException ne) {
        }

        this.nonJtaDataSource = nonJtaDataSource;
    }

    public List<String> getMappingFileNames() {
        return mappingFileNames;
    }

    public void setMappingFileNames(List<String> mappingFileNames) {
        this.mappingFileNames = mappingFileNames;
    }

    public void addMappingFileName(String mappingFileName) {

        mappingFileNames.add(mappingFileName);
    }

    public List<URL> getJarFileUrls() {
        return jarFileUrls;
    }

    public void setJarFileUrls(List<URL> jarFileUrls) {
        this.jarFileUrls = jarFileUrls;
    }

    public void addJarFileUrlPath(String JarFileUrlPath) {
        URL url = null;
        String CanonicalPath = persistenceUnitRootUrl.getPath() + File.separator + JarFileUrlPath;

        try {
            url = new URL(CanonicalPath);
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }

        if (url != null) {
            jarFileUrls.add(url);
        }
    }

    public URL getPersistenceUnitRootUrl() {
        return persistenceUnitRootUrl;
    }

    public void setPersistenceUnitRootUrl(URL persistenceUnitRootUrl) {
        this.persistenceUnitRootUrl = persistenceUnitRootUrl;
    }

    public void setPersistenceUnitRootUrlPath(String UnitRootUrlPath) {
        URL url = null;

        try {
            if (UnitRootUrlPath != null) {
                File root = new File(UnitRootUrlPath);
                url = root.toURL();
            }
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }

        if (url != null) {
            persistenceUnitRootUrl = url;
        }
    }

    public List<String> getManagedClassNames() {
        return managedClassNames;
    }

    public void setManagedClassNames(List<String> managedClassNames) {
        this.managedClassNames = managedClassNames;
    }

    public void addManagedClassName(String managedClassName) {
        managedClassNames.add(managedClassName);
    }

    public boolean excludeUnlistedClasses() {
        return excludeUnlistedClasses;
    }

    public void setExcludeUnlistedClasses(boolean excludeUnlistedClasses) {
        this.excludeUnlistedClasses = excludeUnlistedClasses;
    }

    public Properties getProperties() {
        return properties;
    }

    public void setProperties(Properties properties) {
        this.properties = properties;
    }

    public void addProperty(String name, String value) {
        properties.put(name, value);
    }

    public ClassLoader getClassLoader() {
        return classLoader;
    }

    public void setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
    }

    /**
     * Add a transformer supplied by the provider that will be called for every
     * new class definition or class redefinition that gets loaded by the loader
     * returned by the PersistenceInfo.getClassLoader method. The transformer
     * has no effect on the result returned by the
     * PersistenceInfo.getTempClassLoader method. Classes are only transformed
     * once within the same classloading scope, regardless of how many
     * persistence units they may be a part of.
     */
    public void addTransformer(ClassTransformer transformer) {
        transformers.add(transformer);
    }

    public List<ClassTransformer> getTransformers() {
        return transformers;
    }

    /**
     * Return a new instance of a ClassLoader that the provider may use to
     * temporarily load any classes, resources, or open URLs. The scope and
     * classpath of this loader is exactly the same as that of the loader
     * returned by PersistenceInfo.getClassLoader. None of the classes loaded by
     * this class loader will be visible to application components. The
     * container does not use or maintain references to this class loader after
     * returning it to the provider.
     */
    public ClassLoader getNewTempClassLoader() {
        return newTempClassLoader;
    }

    public void setNewTempClassLoader(ClassLoader newTempClassLoader) {
        this.newTempClassLoader = newTempClassLoader;
    }

}
