/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.metadata;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.xplus.deployer.DatabaseManager;
import org.ow2.xplus.metadata.object.ClassMetaData;
import org.ow2.xplus.metadata.query.NamedNativeQueryMapping;
import org.ow2.xplus.metadata.query.NamedQueryMapping;
import org.ow2.xplus.metadata.query.ResultSetMapping;
import org.ow2.xplus.metadata.relational.RdbIdGenerator;
import org.ow2.xplus.metadata.relational.RdbTable;
import org.ow2.xplus.util.NamingStrategy;

/**
 * @author Lin Liang
 */

public class UnitMetaData {

    private String unitName;

    private String databaseName;

    private String databaseDialect;

    /**
     * The database manager.
     */
    private DatabaseManager database;

    /**
     * all entity class mapping meta data map<the name of the class, class
     * metameta>
     */
    private Map<String, ClassMetaData> classMetaDatas = new HashMap<String, ClassMetaData>();

    /**
     * all relational database table meta data map<the full name of the table,
     * table metameta>
     */
    private Map<String, RdbTable> tables = new HashMap<String, RdbTable>();

    /**
     * all table id generator Map<the name of table, table IdGenerator>
     */
    private Map<String, RdbIdGenerator> tableIdGens = new HashMap<String, RdbIdGenerator>();

    public static String DEFAULT_TABLE_GENERATOR_NAME = "default";

    /**
     * all sequence id generator Map<the name of IdGenerator, sequence
     * IdGenerator>
     */
    private Map<String, RdbIdGenerator> sequenceIdGens = new HashMap<String, RdbIdGenerator>();

    public static String DEFAULT_SEQUENCE_GENERATOR_NAME = "default";

    private Map<String, NamedNativeQueryMapping> namedNativeQuerys = new HashMap<String, NamedNativeQueryMapping>();

    private Map<String, NamedQueryMapping> namedQuerys = new HashMap<String, NamedQueryMapping>();

    private Map<String, ResultSetMapping> resultSets = new HashMap<String, ResultSetMapping>();

    public UnitMetaData(String unitName) {
        this.unitName = unitName;
    }

    /**
     * Used to determine if any MetaData for the entity class
     * @param className the entity class name
     * @return true if has class metadata for this class
     */
    public boolean hasClassMetaData(String className) {
        if (classMetaDatas.get(className) == null)
            return false;
        return true;
    }

    /**
     * get class metadata for the entity class if null, create and return new
     * metadata
     * @param className the entity class name
     * @return the class metadata for this class
     */
    public ClassMetaData getClassMetaData(String className) {
        ClassMetaData classMetaData = classMetaDatas.get(className);
        if (classMetaData == null) {
            classMetaData = new ClassMetaData(unitName, className);
            classMetaDatas.put(className, classMetaData);
        }
        return classMetaData;
    }

    /**
     * get class metadata for the entity class can return null
     * @param className the entity class name
     * @return the class metadata for this class
     */
    public ClassMetaData findClassMetaData(String className) {
        return classMetaDatas.get(className);
    }

    /**
     * get all class metadata in the repository
     * @return all class metadata in the repository
     */
    public List<ClassMetaData> getClassMetaDatas() {
        Collection<ClassMetaData> allClassMetaDatas = classMetaDatas.values();
        List<ClassMetaData> classMetaDatas = new ArrayList<ClassMetaData>();
        classMetaDatas.addAll(allClassMetaDatas);
        return classMetaDatas;
    }

    /**
     * get table infromation with the table name
     * @param tableName the name of the table
     * @return the table with the name
     */
    public RdbTable getTable(String tableName) {
        RdbTable table = tables.get(tableName);
        if (table == null) {
            table = new RdbTable(tableName);
            tables.put(tableName, table);
        }

        return table;
    }

    /**
     * get all tables in the repository
     * @return all tables in the repository
     */
    public List<RdbTable> getTables() {
        Collection<RdbTable> allTables = tables.values();
        List<RdbTable> tables = new ArrayList<RdbTable>();
        tables.addAll(allTables);
        return tables;
    }

    /**
     * add Table IdGenerator to mapping
     * @param tableIdgen Table IdGenerator
     */
    public void addTableIdGen(String primaryTableName, RdbIdGenerator tableIdgen) {
        tableIdGens.put(primaryTableName, tableIdgen);
    }

    /**
     * get Table IdGenerator with the given name
     * @param tableIdGenName Table IdGenerator name
     * @return Table IdGenerator with the given name
     */
    public RdbIdGenerator getTableIdGen(String tableIdGenName) {
        return tableIdGens.get(tableIdGenName);
    }

    /**
     * add Sequence IdGenerator to mapping
     * @param sequenceIdgen Sequence IdGenerator
     */
    public void addSequenceIdGen(RdbIdGenerator sequenceIdgen) {
        sequenceIdGens.put(sequenceIdgen.getName(), sequenceIdgen);
    }

    /**
     * get Sequence IdGenerator with the given name
     * @param sequenceIdGenName Sequence IdGenerator name
     * @return Sequence IdGenerator with the given name
     */
    public RdbIdGenerator getSequenceIdGen(String sequenceIdGenName) {
        return sequenceIdGens.get(sequenceIdGenName);
    }

    public String getUnitName() {
        return unitName;
    }

    public String getDatabaseName() {
        return databaseName;
    }

    public void setDatabaseName(String databaseName) {
        this.databaseName = databaseName;
    }

    public DatabaseManager getDatabase() {
        return database;
    }

    public void setDatabase(DatabaseManager database) {
        this.database = database;
    }

    /**
     * get the Canonical Name of the entity
     * @param name the unqualify name of the entity class name
     * @return the Canonical Name of the entity class
     */
    public String findEntityName(String name) {
        for (String entityName : classMetaDatas.keySet()) {
            if (NamingStrategy.unqualify(entityName).equals(name)) {
                return entityName;
            }
        }
        return null;
    }

    public NamedNativeQueryMapping getNamedNativeQuery(String name) {
        return namedNativeQuerys.get(name);
    }

    public void addNamedNativeQuery(String name, NamedNativeQueryMapping namedNativeQuery) {
        namedNativeQuerys.put(name, namedNativeQuery);
    }

    public NamedQueryMapping getNamedQuery(String name) {
        return namedQuerys.get(name);
    }

    public void addNamedQuery(String name, NamedQueryMapping namedQuery) {
        namedQuerys.put(name, namedQuery);
    }

    public ResultSetMapping getResultSet(String name) {
        return resultSets.get(name);
    }

    public void addResultSet(String name, ResultSetMapping resultSet) {
        resultSets.put(name, resultSet);
    }

    public String getDatabaseDialect() {
        return databaseDialect;
    }

    public void setDatabaseDialect(String databaseType) {
        this.databaseDialect = databaseType;
    }

}
