/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.metadata.object;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.Embeddable;
import javax.persistence.Entity;
import javax.persistence.InheritanceType;
import javax.persistence.MappedSuperclass;

import org.ow2.xplus.metadata.MetaDataRepository;
import org.ow2.xplus.metadata.callback.CallbackMetaData;
import org.ow2.xplus.metadata.strategy.InheritanceStrategy;
import org.ow2.xplus.util.ClassStrategy;
import org.ow2.xplus.util.NamingStrategy;

/**
 * metadata class for defining persistence information on a class.
 * @author Lin Liang
 */

public class ClassMetaData {
    /**
     * the name of entity class
     */
    private String entityName;

    /**
     * if the class is embeded
     */
    private boolean embeded = false;

    /**
     * the name of primary table of the entity
     */
    private String primaryTableName;

    /**
     * The inheritance type to use for the entity inheritance hierarchy.
     */
    private InheritanceType inheritanceType;

    /**
     * inheritance root class
     */
    private Class root = null;

    /**
     * inheritance parent class
     */
    private Class parent = null;

    private boolean isInheritanceRoot = false;

    /**
     * The inheritance strategy to use for the entity inheritance hierarchy.
     */
    private InheritanceStrategy inheritanceStrategy = null;

    private boolean isMappedSuperclass = false;

    /**
     * whether the class has a rdb table mapping
     */
    private boolean hasTable = true;

    /**
     * the name of direct subclasses
     */
    private List<String> subClasseNames = new ArrayList<String>();

    /**
     * all field metadatas Map<the name of field, field metadata>
     */
    private Map<String, FieldMetaData> fields = new HashMap<String, FieldMetaData>();

    /**
     * all attribute overrides map<the name of the property, the name of
     * column>
     */
    private Map<String, String> overrides = new HashMap<String, String>();

    /**
     * the name of persistence unit contains this entity
     */
    private String unitName;

    /**
     * lifecycle callbacks
     */
    private CallbackMetaData callbacks;

    /**
     * the value of the discriminator column for entities of the given type
     */
    private String discriminatorValue;

    /**
     * constructor
     * @param unitName the name of persistence unit contains this entity
     * @param entityName the name of the entity class
     */
    public ClassMetaData(String unitName, String entityName) {
        this.entityName = entityName;
        this.unitName = unitName;
        buildHierarchy(unitName);
    }

    /**
     * build class inheritance hierarchy
     * @param unitName the name of persistence unit contains this entity
     */
    @SuppressWarnings("unchecked")
    private void buildHierarchy(String unitName) {
        // find the inheritance root
        Class vistor = ClassStrategy.forName(entityName);
        while (isPersistence(vistor.getSuperclass())) {
            vistor = vistor.getSuperclass();
        }
        root = vistor;

        if (root.getName().equals(entityName)) {
            isInheritanceRoot = true;
        }
        // if not root, add it to parent subclasses list
        if (!isInheritanceRoot) {
            parent = ClassStrategy.forName(entityName).getSuperclass();
            ClassMetaData parentMetaData = MetaDataRepository.getUnitMetaData(unitName).getClassMetaData(parent.getName());
            parentMetaData.addSubClassName(this.getEntityName());
        }
    }

    /**
     * Used to determine if it's a persistence class
     * @param clazz the entity class
     * @return true if it's a persistence class
     */
    @SuppressWarnings("unchecked")
    private boolean isPersistence(Class clazz) {
        if (clazz != null) {
            if (clazz.isAnnotationPresent(Entity.class) || clazz.isAnnotationPresent(Embeddable.class)
                    || clazz.isAnnotationPresent(MappedSuperclass.class)) {
                return true;
            }
        }
        return false;
    }

    /**
     * @return inheritance root class
     */
    public Class getInheritanceRootClass() {
        return root;
    }

    /**
     * add AttributeOverrides to mapping
     * @param propertyName The name of the property in the embedded object that
     *        is being mapped
     * @param columnName the name of the column that is being mapped to the
     *        persistent attribute.
     */
    public void addAttributeOverrides(String propertyName, String columnName) {
        overrides.put(propertyName, columnName);
    }

    public Map<String, String> getAttributeOverrides() {
        return overrides;
    }

    public FieldMetaData getFieldMetaData(String fieldName) {
        FieldMetaData field = fields.get(fieldName);
        if (field == null) {
            field = new FieldMetaData(fieldName);
            fields.put(fieldName, field);
        }
        return field;
    }

    public FieldMetaData findFieldMetaData(String fieldName) {
        return fields.get(fieldName);
    }

    public String getEntityName() {
        return entityName;
    }

    public void setEntityName(String entityName) {
        this.entityName = entityName;
    }

    public boolean isEmbeded() {
        return embeded;
    }

    public void setEmbeded(boolean embeded) {
        this.embeded = embeded;
    }

    public InheritanceType getInheritanceType() {
        return inheritanceType;
    }

    public void setInheritanceType(InheritanceType inheritanceType) {
        this.inheritanceType = inheritanceType;
    }

    public String getPrimaryTableName() {
        if (primaryTableName == null)
            primaryTableName = NamingStrategy.unqualify(entityName);
        return primaryTableName;
    }

    public void setPrimaryTableName(String tableName) {
        this.primaryTableName = tableName;
    }

    public boolean isInheritanceRoot() {
        return isInheritanceRoot;
    }

    private void addSubClassName(String subClass) {
        subClasseNames.add(subClass);
    }

    public List<String> getSubClasseNames() {
        return subClasseNames;
    }

    /**
     * get all its own fields
     * @return
     */
    public List<FieldMetaData> getAllFields() {
        List<FieldMetaData> allFields = new ArrayList<FieldMetaData>();
        allFields.addAll(fields.values());
        return allFields;
    }

    /**
     * get all fields include inherited from its parents
     * @return
     */
    public List<FieldMetaData> getDeclaredFields() {
        List<FieldMetaData> allFields = new ArrayList<FieldMetaData>();
        allFields.addAll(fields.values());
        if (!isInheritanceRoot()) {
            ClassMetaData parentMetaData = MetaDataRepository.getUnitMetaData(unitName).getClassMetaData(parent.getName());
            allFields.addAll(parentMetaData.getDeclaredFields());
        }
        return allFields;
    }

    public boolean isMappedSuperclass() {
        return isMappedSuperclass;
    }

    public void setMappedSuperclass(boolean isMappedSuperclass) {
        this.isMappedSuperclass = isMappedSuperclass;
    }

    public boolean hasTable() {
        return hasTable;
    }

    public void setHasTable(boolean hasTable) {
        this.hasTable = hasTable;
    }

    public Class getParent() {
        return parent;
    }

    public InheritanceStrategy getInheritanceStrategy() {
        return inheritanceStrategy;
    }

    public void setInheritanceStrategy(InheritanceStrategy inheritanceStrategy) {
        this.inheritanceStrategy = inheritanceStrategy;
    }

    public String getUnitName() {
        return unitName;
    }

    public CallbackMetaData getCallbacks() {
        return callbacks;
    }

    public void setCallbacks(CallbackMetaData callbacks) {
        this.callbacks = callbacks;
    }

    public String getDiscriminatorValue() {
        return discriminatorValue;
    }

    public void setDiscriminatorValue(String discriminatorValue) {
        this.discriminatorValue = discriminatorValue;
    }

}
