/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.metadata.relational;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Lin Liang
 */

public class RdbTable {

    /**
     * the name of the table.
     */
    private String name = null;

    /**
     * the schema of the table
     */
    private String schema = null;

    /**
     * the catalog of the table
     */
    private String catalog = null;

    /**
     * the column of the table Map <column name, index in the column array>
     */
    private Map<String, Integer> columnMap = new HashMap<String, Integer>();

    private List<RdbColumn> columns = new ArrayList<RdbColumn>();

    /**
     * the primary key of the table
     */
    private RdbPrimaryKey primaryKey;

    /**
     * the index of the table Map <index name, index>
     */
    private Map<String, RdbIndex> indexes = new HashMap<String, RdbIndex>();

    /**
     * the relation of the table (Foreign key, relation table) Map<the name of
     * referenced field, relation>
     */
    private Map<String, RdbRelation> relations = new HashMap<String, RdbRelation>();

    /**
     * the unique constraint of the table
     */
    private Map<String, RdbUniqueConstraint> uniqueConstraints = new HashMap<String, RdbUniqueConstraint>();

    /**
     * Map<the name of column field, the info of override column>
     */
    private Map<String, RdbColumn> attributeOverrides = new HashMap<String, RdbColumn>();

    private RdbDiscriminator discriminator;

    private RdbInheritanceJoinedKey inheritanceJoinedKey;

    private List<String> secondaryTableNames = new ArrayList<String>();

    private boolean existed = true;

    public RdbTable() {
    }

    public RdbTable(String name) {
        setName(name);
    }

    public void merge(RdbTable mergeTable) {
        List<RdbColumn> mergeColumns = mergeTable.getColumns();
        for (RdbColumn mergeColumn : mergeColumns) {
            RdbColumn column = new RdbColumn(mergeColumn);
            column.setTableName(name);
            column.setPrimaryKey(mergeColumn.isPrimaryKey());
            addColumn(column);
        }
        primaryKey = mergeTable.primaryKey;
        relations.putAll(mergeTable.relations);
        uniqueConstraints.putAll(mergeTable.uniqueConstraints);
        attributeOverrides.putAll(mergeTable.attributeOverrides);
        secondaryTableNames.addAll(mergeTable.secondaryTableNames);
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        if (name != null)
            this.name = name;
    }

    public String getSchema() {
        return schema;
    }

    public void setSchema(String schema) {
        if (schema != null)
            this.schema = schema;
    }

    public String getCatalog() {
        return catalog;
    }

    public void setCatalog(String catalog) {
        if (catalog != null)
            this.catalog = catalog;
    }

    public String getCanonicalName() {
        String fullName = new String("");
        if (catalog != null) {
            fullName += catalog;
            fullName += ".";
        }
        if (schema != null) {
            fullName += name;
            fullName += ".";
        }
        if (name != null) {
            fullName += name;
        }
        return fullName;
    }

    public RdbColumn getColumn(String name) {
        if (name != null && columnMap != null) {
            if (columnMap.get(name.toUpperCase()) != null) {
                int index = columnMap.get(name.toUpperCase()).intValue();
                return columns.get(index);
            }
        }
        return null;
    }

    public void addColumn(RdbColumn column) {
        RdbColumn old = getColumn(column.getName());
        if (old == null) {
            columnMap.put(column.getName().toUpperCase(), new Integer(columns.size()));
            columns.add(column);
        } else {
            old = column;
        }
    }

    public void removeColumn(String name) {
        if (name != null && columnMap != null) {
            if (columnMap.get(name.toUpperCase()) != null) {
                int index = columnMap.get(name.toUpperCase()).intValue();
                columns.remove(index);
            }
        }
    }

    public RdbPrimaryKey getPrimaryKey() {
        return primaryKey;
    }

    public void setPrimaryKey(RdbPrimaryKey primaryKey) {
        this.primaryKey = primaryKey;
    }

    public RdbIndex getIndex(String indexName) {
        return indexes.get(indexName);
    }

    public RdbIndex addIndex(RdbIndex index) {
        RdbIndex current = indexes.get(index.getName());
        if (current != null) {
            throw new RuntimeException("Index " + index.getName() + " already exists!");
        }
        indexes.put(index.getName(), index);
        return index;
    }

    public void AddUniqueConstraint(RdbUniqueConstraint uc) {
        if (!uniqueConstraints.containsKey(uc.getName())) {
            uniqueConstraints.put(uc.getName(), uc);
        }
    }

    public RdbUniqueConstraint getUniqueConstraint(String name) {
        RdbUniqueConstraint uc = null;
        if (!uniqueConstraints.containsKey(name)) {
            uc = uniqueConstraints.get(name);
        }
        return uc;
    }

    public List<RdbUniqueConstraint> getUniqueConstraints() {
        Collection<RdbUniqueConstraint> allUniqueConstraints = uniqueConstraints.values();
        List<RdbUniqueConstraint> uniqueConstraint = new ArrayList<RdbUniqueConstraint>();
        uniqueConstraint.addAll(allUniqueConstraints);
        return uniqueConstraint;
    }

    public void addRelation(String fieldName, RdbRelation relation) {
        relations.put(fieldName, relation);
    }

    public RdbRelation getRelation(String fieldName) {
        return relations.get(fieldName);
    }

    public List<RdbRelation> getRelations() {
        Collection<RdbRelation> allRelations = relations.values();
        List<RdbRelation> relation = new ArrayList<RdbRelation>();
        relation.addAll(allRelations);
        return relation;
    }

    public void addAttributeOverride(String fieldName, RdbColumn overrrideColumnInfo) {
        attributeOverrides.put(fieldName, overrrideColumnInfo);
    }

    /**
     * return Attribute Override info, if there is no Attribute Override, return
     * null
     * @param columnName the name of the column
     * @return
     */
    public RdbColumn getAttributeOverride(String fieldName) {
        return attributeOverrides.get(fieldName);
    }

    public List<RdbColumn> getColumns() {
        return columns;
    }

    public RdbDiscriminator getDiscriminator() {
        return discriminator;
    }

    public void setDiscriminator(RdbDiscriminator discriminator) {
        this.discriminator = discriminator;
    }

    public RdbInheritanceJoinedKey getInheritanceJoinedKey() {
        return inheritanceJoinedKey;
    }

    public void setInheritanceJoinedKey(RdbInheritanceJoinedKey inheritanceJoinedKey) {
        this.inheritanceJoinedKey = inheritanceJoinedKey;
    }

    public List<String> getSecondaryTableNames() {
        return secondaryTableNames;
    }

    public void addSecondaryTableName(String secondaryTableName) {
        secondaryTableNames.add(secondaryTableName);
    }

    public boolean isExisted() {
        return existed;
    }

    public void setExisted(boolean existed) {
        this.existed = existed;
    }

}
