/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.operator.accessor;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.persistence.CascadeType;

import org.jdom.Element;

import org.ow2.xplus.exception.PropertyAccessException;
import org.ow2.xplus.metadata.MetaDataRepository;
import org.ow2.xplus.metadata.UnitMetaData;
import org.ow2.xplus.metadata.object.ClassMetaData;
import org.ow2.xplus.metadata.object.FieldMetaData;
import org.ow2.xplus.metadata.relational.RdbColumn;
import org.ow2.xplus.util.NamingStrategy;

/**
 * abstract implement of the property accessor
 * @author Lin Liang
 */

public abstract class AbstractAccessor implements Accessor {

    protected ClassMetaData classMetaData;

    protected FieldMetaData field;

    /**
     * the mapping metadata of the unit
     */
    protected UnitMetaData unit;

    /**
     * the columns in the table of this class
     */
    protected List<RdbColumn> columns = new ArrayList<RdbColumn>();

    /**
     * the value in the columns
     */
    protected List values = new ArrayList();

    protected CascadeType needCascadeType;

    public AbstractAccessor(ClassMetaData classMetaData, FieldMetaData field) {
        this.unit = MetaDataRepository.getUnitMetaData(classMetaData.getUnitName());
        this.classMetaData = classMetaData;
        this.field = field;
    }

    /**
     * fetch the value of the given property in the entity
     * @param entityData the entity instance
     * @param propertyName the name of the property
     * @return the value of the given property in the entity
     */
    protected Object getFieldValue(Object entityData, String propertyName) {
        Method[] methods = entityData.getClass().getMethods();
        Object value = null;

        String getMethodName = NamingStrategy.getFieldAccessMethod(propertyName, "get");
        String isMethodName = NamingStrategy.getFieldAccessMethod(propertyName, "is");
        for (Method method : methods) {
            if (method.getName().equals(getMethodName) || method.getName().equals(isMethodName)) {
                try {
                    value = method.invoke(entityData, new Object[] {});
                } catch (IllegalArgumentException iae) {
                    PropertyAccessException ex = new PropertyAccessException(iae, PropertyAccessException.ILLEGAL_ARGUMENT_ERROR);
                    ex.showMessage();
                } catch (IllegalAccessException iace) {
                    PropertyAccessException ex = new PropertyAccessException(iace, PropertyAccessException.ILLEGAL_ACCESS_ERROR);
                    ex.showMessage();
                } catch (InvocationTargetException ite) {
                    PropertyAccessException ex = new PropertyAccessException(ite, PropertyAccessException.INVOCATION_TARGET_ERROR);
                    ex.showMessage();
                }
            }
        }

        if (field.getValue().getValueStrategy() == null)
            return value;
        return field.getValue().getValueStrategy().getDataValue(value);
    }

    /**
     * override the column info based on
     * @AttributeOverride annotation parsed before
     * @param accessor the property accessor
     */
    protected void overrideColumn(Accessor accessor) {
        List<RdbColumn> originColumns = accessor.getColumns();

        for (RdbColumn column : originColumns) {
            String embeddedFieldName = column.getCanonicalFieldName();
            RdbColumn overrideColumn = field.getValue().getAttributeOverride(embeddedFieldName);
            if (overrideColumn != null) {
                if (!overrideColumn.getName().equals(""))
                    column.setName(overrideColumn.getName());
                if (!overrideColumn.getTableName().equals(""))
                    column.setTableName(overrideColumn.getTableName());
            }

            overrideColumn = unit.getTable(classMetaData.getPrimaryTableName()).getAttributeOverride(embeddedFieldName);
            if (overrideColumn != null) {
                if (!overrideColumn.getName().equals(""))
                    column.setName(overrideColumn.getName());
                if (!overrideColumn.getTableName().equals(""))
                    column.setTableName(overrideColumn.getTableName());
            }
        }
    }

    /**
     * get data form object, write it in xml format
     * @param root the root of xml doc
     * @param data the data
     * @param generateid whether to generate id while reading
     */
    public void generateXMLData(Element root, Object data, boolean generateid) {
        if (data == null)
            return;
        obtainValue(data);
        int valuesSize = values.size();
        for (int index = 0; index < columns.size(); ++index) {
            RdbColumn column = columns.get(index);
            if (column.isAutoIncrement())
                continue;
            Object columnData = null;
            if (index < valuesSize)
                columnData = values.get(index);
            else
                break;
            String tableName = column.getTableName();
            Element tableElem = root.getChild(tableName);
            tableElem.setAttribute(column.getName(), columnData.toString());
        }
    }

    /**
     * get cascade entities infomation, write it in xml format
     */
    public void generateCascadeXMLData(Element root, Object data, boolean generateid) {
        return;
    }

    /**
     * the cascade type, used to determine whether to generate cascade data
     */
    public void setNeedCascadeType(CascadeType cascadeType) {
        this.needCascadeType = cascadeType;
    }

    /**
     * @return the columns representing this property
     */
    public List<RdbColumn> getColumns() {
        return columns;
    }

    /**
     * @return the values stored in the columns representing this property
     */
    public List getValues() {
        return values;
    }

    /**
     * the name of tables affected while accessing
     */
    public Set<String> getAffectTableNames(boolean cascade) {
        Set<String> affectTableNames = new HashSet<String>();
        return affectTableNames;
    }

    /**
     * set the value of the field
     * @param entityData the data of the entity
     * @param propertyName the name of the property
     * @param value the value to store
     */
    protected void setFieldValue(Object entityData, String propertyName, Object value) {
        Method[] methods = entityData.getClass().getMethods();

        String setMethodName = NamingStrategy.getFieldAccessMethod(propertyName, "set");
        for (Method method : methods) {
            if (method.getName().equals(setMethodName)) {
                try {
                    Object data = null;
                    if (field.getValue().getValueStrategy() != null)
                        data = field.getValue().getValueStrategy().getObjectValue(value);
                    data = getMatchValue(method.getParameterTypes()[0], data);
                    if (data != null)
                        method.invoke(entityData, new Object[] { data });
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
            }
        }
    }

    /**
     * added by yf
     * @param entityData
     * @param propertyName
     * @param value
     */
    protected void setFieldValue4Copy(Object entityData, String propertyName, Object value){
    	   Method[] methods = entityData.getClass().getMethods();

           String setMethodName = NamingStrategy.getFieldAccessMethod(propertyName, "set");
           for (Method method : methods) {
               if (method.getName().equals(setMethodName)) {
                   try {
                           method.invoke(entityData, new Object[] { value });
                   } catch (Exception ex) {
                       ex.printStackTrace();
                   }
               }
           }
    }
    
   
    private Object getMatchValue(Class expectType, Object originValue) {
        Object value = originValue;
        if (originValue instanceof BigDecimal) {
            value = new Integer(((BigDecimal) originValue).intValue());
        }
        return value;
    }

    /**
     * get the identity of the entity
     * 
     * @param entityData the data of the entity
     * @param classMetaData the metadata of the entity
     * @return the identity of the entity
     */
    protected Object getIdValue(Object entityData, ClassMetaData classMetaData) {
        for (FieldMetaData field : classMetaData.getAllFields()) {
            if (field.isId()) {
                Object key = getFieldValue(entityData, field.getFieldName());
                return key;
            }
        }

        return null;
    }

}
