/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.operator.accessor;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.ow2.xplus.exception.PropertyAccessException;
import org.ow2.xplus.metadata.object.ClassMetaData;
import org.ow2.xplus.metadata.object.FieldMetaData;
import org.ow2.xplus.util.ClassStrategy;
import org.ow2.xplus.util.NamingStrategy;

/**
 * abstract implement of the visitor
 * @author Lin Liang
 */

public abstract class AbstractVisitor implements Visitor {

    /**
     * get the value of the field in the entity
     * @param entityData the entity data
     * @param propertyName the name of the property
     * @return the value of the field in the entity
     */
    protected Object getFieldValue(Object entityData, String propertyName) {
        Object fieldValue = null;

        Method[] methods = entityData.getClass().getMethods();

        String getMethodName = NamingStrategy.getFieldAccessMethod(propertyName, "get");
        for (Method method : methods) {
            if (method.getName().equals(getMethodName)) {
                try {
                    fieldValue = method.invoke(entityData, new Object[] {});
                } catch (IllegalArgumentException iae) {
                    PropertyAccessException ex = new PropertyAccessException(iae, PropertyAccessException.ILLEGAL_ARGUMENT_ERROR);
                    ex.showMessage();
                } catch (IllegalAccessException iace) {
                    PropertyAccessException ex = new PropertyAccessException(iace, PropertyAccessException.ILLEGAL_ACCESS_ERROR);
                    ex.showMessage();
                } catch (InvocationTargetException ite) {
                    PropertyAccessException ex = new PropertyAccessException(ite, PropertyAccessException.INVOCATION_TARGET_ERROR);
                    ex.showMessage();
                }
            }
        }

        if (fieldValue == null) {
            try {
                Class type = entityData.getClass().getDeclaredField(propertyName).getType();
                fieldValue = ClassStrategy.newInstance(type);
            } catch (SecurityException e) {
                throw new PropertyAccessException(e, PropertyAccessException.SECURITY_ERROR);
            } catch (NoSuchFieldException e) {
                throw new PropertyAccessException(e, PropertyAccessException.NO_SUCH_FIELD_ERROR);
            } catch (InstantiationException e) {
                throw new PropertyAccessException(e, PropertyAccessException.INSTANTIATION_ERROR);
            } catch (IllegalAccessException e) {
                throw new PropertyAccessException(e, PropertyAccessException.ILLEGAL_ACCESS_ERROR);
            }
        }

        return fieldValue;
    }

    /**
     * store the value to the field of the entity
     * @param entityData the entity data
     * @param propertyName the name of the property
     * @param value the value to store
     */
    protected void setFieldValue(Object entityData, String propertyName, Object value) {
        Method[] methods = entityData.getClass().getMethods();

        String getMethodName = NamingStrategy.getFieldAccessMethod(propertyName, "set");
        for (Method method : methods) {
            if (method.getName().equals(getMethodName)) {
                try {
                    method.invoke(entityData, new Object[] { value });
                } catch (IllegalArgumentException iae) {
                    PropertyAccessException ex = new PropertyAccessException(iae, PropertyAccessException.ILLEGAL_ARGUMENT_ERROR);
                    ex.showMessage();
                } catch (IllegalAccessException iace) {
                    PropertyAccessException ex = new PropertyAccessException(iace, PropertyAccessException.ILLEGAL_ACCESS_ERROR);
                    ex.showMessage();
                } catch (InvocationTargetException ite) {
                    PropertyAccessException ex = new PropertyAccessException(ite, PropertyAccessException.INVOCATION_TARGET_ERROR);
                    ex.showMessage();
                }
            }
        }
    }

    /**
     * get the identity of the entity
     * 
     * @param entityData the entity data
     * @param classMetaData the metadata of the entity
     * @return the identity of the entity
     */
    protected Object getIdValue(Object entityData, ClassMetaData classMetaData) {
        for (FieldMetaData field : classMetaData.getAllFields()) {
            if (field.isId()) {
                Object key = getFieldValue(entityData, field.getFieldName());
                return key;
            }
        }

        return null;
    }

}
