/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.query;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityNotFoundException;
import javax.persistence.FlushModeType;
import javax.persistence.NonUniqueResultException;
import javax.persistence.Query;
import javax.persistence.TemporalType;

import org.apache.commons.beanutils.DynaBean;
import org.apache.commons.beanutils.DynaProperty;
import org.apache.ddlutils.platform.ModelBasedResultSetIterator;

import org.ow2.xplus.broker.PersistenceManager;
import org.ow2.xplus.context.PersistenceContext;
import org.ow2.xplus.deployer.DatabaseManager;
import org.ow2.xplus.metadata.UnitMetaData;
import org.ow2.xplus.metadata.object.ClassMetaData;
import org.ow2.xplus.metadata.relational.RdbColumn;
import org.ow2.xplus.metadata.relational.RdbRelation;
import org.ow2.xplus.metadata.relational.RdbTable;
import org.ow2.xplus.operator.EntityCreator;
import org.ow2.xplus.operator.Finder;
import org.ow2.xplus.util.ClassStrategy;

/**
 * the abstract implement of Query interface
 * @author Lin Liang
 */

public abstract class AbstractQuery implements Query {

    private FlushModeType flushMode;

    private int maxResult;

    private int startPosition;

    private UnitMetaData unit;

    protected String sqlString;

    protected PersistenceManager pm;

    /**
     * The database manager.
     */
    protected DatabaseManager database;

    private PersistenceContext context;

    private Map<String, String> columnsToTypes = new HashMap<String, String>();

    /** Maps column names to properties. */
    private Map<String, String> columnsToProperties = new HashMap<String, String>();

    public AbstractQuery(String queryString, PersistenceManager pm) {
        this.pm = pm;
        if (pm != null) {
            context = pm.getContext();
            unit = pm.getUnitMetaData();
            database = unit.getDatabase();
        }
    }

    /**
     * Performs the given SQL query.
     * @param model The database model to use
     * @param sql The sql query to perform
     */
    protected int executeQuery() {
        return database.getPlatform().execute(database.getModel(), sqlString);
    }

    /**
     * Execute the query and return the query results as a List.
     * @param type the type of the entity
     * @return a list of the results
     */
    @SuppressWarnings("unchecked")
    protected List getQueryResults(Class type) {
        List results = new ArrayList();

        // get all columns of the entity
        ClassMetaData entityMetaData = unit.findClassMetaData(type.getName());

        if (entityMetaData != null) {
            String tableName = entityMetaData.getPrimaryTableName();
            RdbTable table = unit.getTable(tableName);

            for (RdbRelation relation : table.getRelations()) {
                String referencedTableName = relation.getReferencedTableName();
                RdbTable referencedTable = unit.getTable(referencedTableName);
                for (RdbColumn column : referencedTable.getColumns()) {
                    columnsToTypes.put(column.getName(), ClassStrategy.toName(column.getJavaType()));
                }
            }

            for (RdbColumn column : table.getColumns()) {
                columnsToTypes.put(column.getName(), ClassStrategy.toName(column.getJavaType()));
            }
        }

        // execute the query
        ModelBasedResultSetIterator resultIter = (ModelBasedResultSetIterator) database.getPlatform().query(database.getModel(),
                sqlString);

        resultIter.getColumnsToTypes().putAll(columnsToTypes);
        resultIter.getColumnsToProperties().putAll(columnsToProperties);

        if (type != null) {
            for (ClassMetaData metaData : unit.getClassMetaDatas()) {
                if (type.getName().equals(metaData.getEntityName())) {
                    EntityCreator creator = new EntityCreator(unit, context);
                    Finder finder = new Finder(creator);
                    results = finder.get(type, resultIter);
                    return results;
                }
            }
        }

        while (resultIter.hasNext()) {
            DynaBean bean = (DynaBean) resultIter.next();
            DynaProperty[] props = bean.getDynaClass().getDynaProperties();
            // String propName = columnMapping.get(props[0].getName());
            // if (propName == null)
            results.add(bean.get(props[0].getName()));
        }
        return results;
    }

    /**
     * Execute a query that returns a single result.
     * @return the result
     * @throws EntityNotFoundException if there is no result
     * @throws NonUniqueResultException if more than one result
     */
    public Object getSingleResult() {
        int size = getResultList().size();
        if (size == 0)
            throw new EntityNotFoundException();
        if (size > 1)
            throw new NonUniqueResultException();
        return getResultList().get(0);
    }

    /**
     * Set the maximum number of results to retrieve.
     * @param maxResult
     * @return the same query instance
     * @throws IllegalArgumentException if argument is negative
     */
    public Query setMaxResults(int maxResult) {
        if (maxResult < 0)
            throw new IllegalArgumentException();
        this.maxResult = maxResult;
        return this;
    }

    /**
     * Set the position of the first result to retrieve.
     * @param startPosition position of the first result, numbered from 0
     * @return the same query instance
     * @throws IllegalArgumentException if argument is negative
     */
    public Query setFirstResult(int startPosition) {
        if (startPosition < 0)
            throw new IllegalArgumentException();
        this.startPosition = startPosition;
        return this;
    }

    /**
     * Set an implementation-specific hint. If the hint name is not recognized,
     * it is silently ignored.
     * @param hintName
     * @param value
     * @return the same query instance
     * @throws IllegalArgumentException if the second argument is not valid for
     *         the implementation
     */
    public Query setHint(String hintName, Object value) {
        return this;
    }

    /**
     * Bind an instance of java.util.Date to a named parameter.
     * @param name
     * @param value
     * @param temporalType
     * @return the same query instance
     * @throws IllegalArgumentException if parameter name does not correspond to
     *         parameter in query string
     */
    public Query setParameter(String name, Date value, TemporalType temporalType) {
        return setParameter(name, value);
    }

    /**
     * Bind an instance of java.util.Calendar to a named parameter.
     * @param name
     * @param value
     * @param temporalType
     * @return the same query instance
     * @throws IllegalArgumentException if parameter name does not correspond to
     *         parameter in query string
     */
    public Query setParameter(String name, Calendar value, TemporalType temporalType) {
        return setParameter(name, value);
    }

    /**
     * Bind an instance of java.util.Date to a positional parameter.
     * @param position
     * @param value
     * @param temporalType
     * @return the same query instance
     * @throws IllegalArgumentException if position does not correspond to
     *         positional parameter of query
     */
    public Query setParameter(int position, Date value, TemporalType temporalType) {
        return setParameter(position, value);
    }

    /**
     * Bind an instance of java.util.Calendar to a positional parameter.
     * @param position
     * @param value
     * @param temporalType
     * @return the same query instance } correspond to positional parameter of
     *         query
     */
    public Query setParameter(int position, Calendar value, TemporalType temporalType) {
        return setParameter(position, value);
    }

    /**
     * Set the flush mode type to be used for the query execution.
     * @param flushMode
     */
    public Query setFlushMode(FlushModeType flushMode) {
        this.flushMode = flushMode;
        return this;
    }

    /**
     * Get the flush mode type to be used for the query execution.
     */
    public FlushModeType getFlushMode() {
        return flushMode;
    }

    /**
     * Get the maximum number of results to retrieve.
     */
    public int getMaxResult() {
        return maxResult;
    }

    public int getStartPosition() {
        return startPosition;
    }

    public Map<String, String> getColumnsToTypes() {
        return columnsToTypes;
    }

    public Map<String, String> getColumnsToProperties() {
        return columnsToProperties;
    }

    public void setUnit(UnitMetaData unit) {
        this.unit = unit;
        database = unit.getDatabase();
    }

}
