/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.query.ejbql.tree;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author Lin Liang
 */

public class FromNode extends AbstractNode {

    private List<Node> decls = new ArrayList<Node>();

    /**
     * id per table name
     */
    private Map<String, String> declarations = new LinkedHashMap<String, String>();

    private Map<String, String> reversedDeclarations = new LinkedHashMap<String, String>();

    /**
     * id per table name
     */
    private Map<String, String> collectionMemberDeclarations = new LinkedHashMap<String, String>();

    /**
     * join declations per id
     */
    private Map<String, List<String>> joinDeclarations = new LinkedHashMap<String, List<String>>();

    /**
     * table ids per join table id
     */
    private Map<String, Set<String>> relations = new LinkedHashMap<String, Set<String>>();

    private List<String> orderedDeclarations = new LinkedList<String>();

    private List<String> removedJoinDeclarations = new LinkedList<String>();

    private String firstTableId = null;

    public FromNode(List args) {
        super(args);

        for (Object decl : args) {
            decls.add((Node) decl);
        }
    }

    public String addDeclaration(String tableName, String ident) {
        if (getRoot().getTree().getIdType(ident) != null)
            return ident;
        String id = declarations.get(tableName);
        if (id != null)
            return id;
        declarations.put(tableName, ident);
        reversedDeclarations.put(ident, tableName);
        return ident;
    }

    public void saveRelation(String tableId, String joinTableId) {
        Set<String> tableIds = relations.get(joinTableId);
        if (tableIds == null)
            tableIds = new HashSet<String>();
        tableIds.add(tableId);
        relations.put(joinTableId, tableIds);
    }

    public void addJoinDeclaration(String ident, String join) {
        // if (getRoot().getTree().getIdType(ident) != null) return;
        List<String> joins = joinDeclarations.get(ident);
        if (joins == null)
            joins = new ArrayList<String>();
        joins.add(join);
        joinDeclarations.put(ident, joins);
    }

    public void addCollectionMemberDeclaration(String tableName, String ident) {
        collectionMemberDeclarations.put(tableName, ident);
    }

    public String getTableId(String tableName) {
        String ident = declarations.get(tableName);
        if (ident == null) {
            ident = getRoot().getTree().getTableId(tableName);
        }
        return ident;
    }

    public String getTableName(String tableId) {
        String ident = reversedDeclarations.get(tableId);
        return ident;
    }

    @Override
    public void process() {
        if (processed)
            return;

        for (Node decl : decls) {
            decl.setRoot(getRoot());
            decl.process();
        }

        super.process();
    }

    private void sortDeclarations() {
        orderedDeclarations.clear();
        String firstDecl = decls.get(0).toSQLString();
        firstTableId = getRoot().getTree().getDialect().getIdFromDeclaration(firstDecl);
        // String firstTableName =
        // getRoot().getTree().getDialect().getNameFromDeclaration(firstDecl);
        orderedDeclarations.add(firstTableId);
        visitDeclarations(firstTableId);
        orderedDeclarations.remove(firstTableId);
    }

    private void visitDeclarations(String joinTableId) {
        Set<String> tableIds = relations.get(joinTableId);
        if (tableIds == null)
            return;
        List<String> needToVisited = new ArrayList<String>();

        for (String tableId : tableIds) {
            if (!orderedDeclarations.contains(tableId)) {
                orderedDeclarations.add(tableId);
                needToVisited.add(tableId);
            } else {
                if (!tableId.equals(joinTableId))
                    removeCircleRelations(tableId, joinTableId);
            }
        }

        for (String tableId : needToVisited) {
            visitDeclarations(tableId);
        }
    }

    public List<String> getJoinDeclaration(String tableId) {
        return joinDeclarations.get(tableId);
    }

    public void removeJoinDeclaration(String tableId) {
        removedJoinDeclarations.add(tableId);
    }

    private void removeCircleRelations(String tableId, String joinTableId) {
        List<String> joins = joinDeclarations.get(tableId);
        List<String> circleJoins = new ArrayList<String>();
        for (String join : joins) {
            if (join.indexOf(joinTableId) != -1) {
                circleJoins.add(join);
            }
        }
        joins.removeAll(circleJoins);
        if (joins.size() == 0)
            joins = null;
    }

    public String toSQLString() {
        sortDeclarations();

        getRoot().getTree().getTableIdDeclarations().putAll(reversedDeclarations);

        StringBuilder from = new StringBuilder(1024);

        from.append("FROM ");

        if (collectionMemberDeclarations.keySet().size() > 0) {
            String tableName = (String) collectionMemberDeclarations.keySet().toArray()[0];
            String tableId = collectionMemberDeclarations.get(tableName);
            from.append(getRoot().getTree().getDialect().getDeclaration(tableName, tableId));
            from.append(" LEFT OUTER JOIN ");
            from.append(decls.get(0).toSQLString());
            from.append(" ON ");
            from.append(joinDeclarations.get(tableId).get(0));
            joinDeclarations.remove(tableId);
            orderedDeclarations.remove(tableId);
        } else {
            from.append(decls.get(0).toSQLString());
        }

        for (String tableId : removedJoinDeclarations) {
            joinDeclarations.remove(tableId);
        }

        for (String id : orderedDeclarations) {
            List<String> joins = joinDeclarations.get(id);
            if (joins == null || joins.size() == 0)
                continue;

            String tableName = reversedDeclarations.get(id);
            if (tableName == null) {
                tableName = getRoot().getTree().getUnit().findClassMetaData(getRoot().getTree().getIdType(id).getCanonicalName())
                        .getPrimaryTableName();
            }
            String referName = getRoot().getTree().getDialect().getDeclaration(tableName, id);
            if (id.equals(firstTableId))
                continue;
            from.append(" LEFT OUTER JOIN ");
            from.append(referName);
            from.append(" ON ");

            from.append(joins.get(0));
            for (int index = 1; index < joins.size(); ++index) {
                from.append(" AND ");
                from.append(joins.get(index));
            }
        }

        return from.toString();
    }

    public Map<String, String> getIdDeclarations() {
        return reversedDeclarations;
    }

}
