/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.query.ejbql.tree;

import java.util.ArrayList;
import java.util.List;

import org.ow2.xplus.metadata.UnitMetaData;
import org.ow2.xplus.metadata.object.ClassMetaData;
import org.ow2.xplus.metadata.relational.RdbColumn;
import org.ow2.xplus.metadata.relational.RdbForeignKey;
import org.ow2.xplus.metadata.relational.RdbRelation;
import org.ow2.xplus.metadata.relational.RdbRelationTable;
import org.ow2.xplus.metadata.relational.RdbTable;
import org.ow2.xplus.query.utils.NameGenerator;
import org.ow2.xplus.util.ClassStrategy;

/**
 * @author Lin Liang
 */

public class ObjectExprNode extends AbstractNode implements ConstructorNode {

    private IdVarNode idVar;

    private List<String> selectColumns = new ArrayList<String>();

    private UnitMetaData unit;

    private Class type;

    public ObjectExprNode(List args) {
        super(args);

        idVar = (IdVarNode) args.get(0);
    }

    @Override
    public void process() {
        if (processed)
            return;

        idVar.setRoot(getRoot());
        idVar.process();

        unit = getRoot().getTree().getUnit();
        type = idVar.getIdType();

        addColumns(type, idVar.toSQLString(), true);

        super.process();
    }

    private void addColumns(Class idType, String idName, boolean cascade) {
        ClassMetaData idMetaData = unit.findClassMetaData(idType.getCanonicalName());
        RdbTable table = unit.getTable(idMetaData.getPrimaryTableName());
        addColumns(table, idName, true);
    }

    private void addColumns(RdbTable table, String idName, boolean cascade) {
        idName = getRoot().getFrom().addDeclaration(table.getName(), idName);
        for (RdbColumn column : table.getColumns()) {
            selectColumns.add(idName + "." + getRoot().getTree().getDialect().getColumnName(column.getName()));
        }

        for (String secTableName : table.getSecondaryTableNames()) {
            RdbTable secTable = unit.getTable(secTableName);
            String secTableId = NameGenerator.getTableName();
            addColumns(secTable, secTableId, true);
        }

        for (RdbRelation relation : table.getRelations()) {
            Class referencedEntity = ClassStrategy.forName(relation.getReferencedEntityName());
            ClassMetaData referencedEntityMetaData = unit.findClassMetaData(referencedEntity.getCanonicalName());
            String referencedEntityId = getRoot().getFrom().getTableId(referencedEntityMetaData.getPrimaryTableName());

            if (referencedEntityId == null) {
                referencedEntityId = NameGenerator.getTableName();
                addColumns(referencedEntity, referencedEntityId, true);
            }

            // if (relation.isInverse()) return;
            referencedEntityId = getRoot().getFrom().addDeclaration(referencedEntityMetaData.getPrimaryTableName(),
                    referencedEntityId);

            String referencedEntityTableName = referencedEntityMetaData.getPrimaryTableName();

            if (relation instanceof RdbForeignKey) {
                processForeignKey((RdbForeignKey) relation, idName, referencedEntityId, referencedEntityTableName);
            } else {
                processRelationTable((RdbRelationTable) relation, idName, referencedEntityId, referencedEntityTableName);
            }
        }
    }

    private void processForeignKey(RdbForeignKey foreignKey, String ownerName, String targetName, String targetTableName) {
        List<String> joinColumnNames = foreignKey.getJoinColumnNames();
        List<String> referencedColumnNames = foreignKey.getReferencedColumnNames();

        getRoot().getFrom().saveRelation(targetName, ownerName);

        FromNode from = getRoot().getFrom();
        for (int index = 0; index < joinColumnNames.size(); ++index) {
            String condition = buildJoinCondition(ownerName, joinColumnNames.get(index), targetName, referencedColumnNames
                    .get(index));
            from.addJoinDeclaration(targetName, condition);
        }
    }

    private void processRelationTable(RdbRelationTable relation, String ownerName, String targetName, String targetTableName) {
        RdbTable relationTable = relation.getRelationTable();
        String relationTableName = NameGenerator.getTableName();
        FromNode from = getRoot().getFrom();
        relationTableName = from.addDeclaration(relationTable.getName(), relationTableName);

        List<String> sourceRelationColumnNames = relation.getSourceRelationColumnNames();
        List<String> sourceColumnNames = relation.getSourceColumnNames();

        getRoot().getFrom().saveRelation(relationTableName, ownerName);
        for (int index = 0; index < sourceRelationColumnNames.size(); ++index) {
            String condition = buildJoinCondition(ownerName, sourceColumnNames.get(index), relationTableName,
                    sourceRelationColumnNames.get(index));
            from.addJoinDeclaration(relationTableName, condition);
        }

        getRoot().getFrom().saveRelation(targetName, relationTableName);
        List<String> targetRelationColumnNames = relation.getTargetRelationColumnNames();
        List<String> targetColumnNames = relation.getTargetColumnNames();
        for (int index = 0; index < targetRelationColumnNames.size(); ++index) {
            String condition = buildJoinCondition(relationTableName, targetRelationColumnNames.get(index), targetName,
                    targetColumnNames.get(index));
            from.addJoinDeclaration(targetName, condition);
        }
    }

    private String buildJoinCondition(String table1, String column1, String table2, String column2) {
        StringBuilder cond = new StringBuilder(256);
        cond.append(table1).append(".").append(column1).append(" = ").append(table2).append(".").append(column2);
        return cond.toString();
    }

    public String toSQLString() {
        StringBuilder object = new StringBuilder(1024);

        object.append(selectColumns.get(0));

        for (int index = 1; index < selectColumns.size(); ++index) {
            object.append(", ");
            object.append(selectColumns.get(index));
        }

        return object.toString();
    }

    public Class getType() {
        return type;
    }

}
