/* -------------------------------------------------------------------------*
 * PKUAS: Peking University Application Server
 * Copyright (C) 2006 Peking University
 *
 * This software is free; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * either version 2.1 or any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * ------------------------------------------------------------------------*/

package org.ow2.xplus.setting;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;

import org.ow2.xplus.logger.LogFactory;
import org.ow2.xplus.logger.Logger;

/**
 * persistence layer setting
 * @author Lin Liang
 */

public class Setting {

    private static String CONFIG_FILE_NAME = "persistence.conf.xml";

    public static String PERSIST_HOME = "home";

    public static String DEPLOY_DIR = "deploydir";

    public static String APP_DIR = "appdir";

    private Properties properties = new Properties();

    private static Logger logger = LogFactory.getLogger(Setting.class);

    public Setting() {
        try {
            readProperties();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void setupDirectories() {
        try {
            setupHomeDir();
            String homeDirPath = (String) properties.get(PERSIST_HOME);

            setupDeployDir(homeDirPath);
            setupAppDir(homeDirPath);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * read properties from config file
     * @throws IOException
     */
    private void readProperties() throws IOException {
        String confPath = System.getProperty("ConfigHome");
        
        if (confPath == null) {
            confPath = new File(".").getCanonicalPath();
        }
        File confFile = new File(confPath, CONFIG_FILE_NAME);

        if (!confFile.exists()) {
            logger.debug("error in finding configure file in path: " + confPath);
            return;
        }
        InputStream is = new FileInputStream(confFile);

        ConfigParser handler = new ConfigParser();
        SAXParserFactory factory = SAXParserFactory.newInstance();
        factory.setNamespaceAware(false);
        factory.setValidating(false);

        SAXParser parser = null;
        try {
            parser = factory.newSAXParser();
        } catch (ParserConfigurationException e) {
            e.printStackTrace();
        } catch (SAXException e) {
            e.printStackTrace();
        }

        if (parser == null)
            return;

        try {
            parser.parse(is, handler);
            properties = handler.getProps();
        } catch (SAXException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            factory = null;
            parser = null;
            handler = null;
        }

        is.close();

    }

    /**
     * Setup persistence home directory
     * @throws IOException
     */
    private void setupHomeDir() throws IOException {

        String homeDirPath = (String) properties.get(PERSIST_HOME);

        if (homeDirPath == null) {
            homeDirPath = new File(".").getCanonicalPath();
        }
        if (homeDirPath.startsWith(".")) {
            homeDirPath = new File(".").getCanonicalPath() + homeDirPath;
        }

        File homeDir = new File(homeDirPath);

        if (homeDir == null || !homeDir.exists()) {
            logger.error("error in checking persistence home directory : " + homeDirPath);
            System.exit(0);
        }

        properties.setProperty(PERSIST_HOME, homeDirPath);
    }

    /**
     * Setup persistence deployment directory
     * @throws IOException
     */
    private void setupDeployDir(String homeDirPath) throws IOException {

        String deployDirPath = (String) properties.get(DEPLOY_DIR);
        File deployDir = null;

        if (deployDirPath == null) {
            deployDir = new File(homeDirPath, "deploy");
        } else if (deployDirPath.startsWith(".")) {
            deployDirPath = homeDirPath + File.separator + deployDirPath;
            deployDir = new File(deployDirPath);
        } else {
            deployDir = new File(deployDirPath);
        }

        if (deployDir == null) {
            logger.debug("error in checking persistence deployment directory.");
            System.exit(0);
        }

        deployDirPath = deployDir.getCanonicalPath();
        if (!deployDir.exists()) {
            logger.debug("persistence deployment directory not found: " + deployDirPath);
            deployDir.mkdir();
        }

        properties.setProperty(DEPLOY_DIR, deployDirPath);
    }

    /**
     * Setup persistence application directory
     * @throws IOException
     */
    private void setupAppDir(String homeDirPath) throws IOException {

        String appDirPath = (String) properties.get(APP_DIR);
        File appDir = null;

        if (appDirPath == null) {
            appDir = new File(homeDirPath, "application");
        } else if (appDirPath.startsWith(".")) {
            appDirPath = homeDirPath + File.separator + appDirPath;
            appDir = new File(appDirPath);
        } else {
            appDir = new File(appDirPath);
        }

        appDirPath = appDir.getCanonicalPath();
        if (appDir == null || !appDir.exists()) {
            logger.error("error in checking persistence application directory: " + appDirPath);
            System.exit(0);
        }

        properties.setProperty(APP_DIR, appDirPath);
    }

    public Properties getProperties() {
        return properties;
    }
    


}
